import os
import json
import pychemia
import numpy as np
from .convergence import ConvergenceKPointGrid
from .relax import IonRelaxation
from ...tasks import Task

__author__ = 'Guillermo Avendano-Franco'


class IdealStrength(Task):
    def __init__(self, structure, workdir='.', executable='vasp', ini_factor=1.0, fin_factor=1.2, delta_factor=0.01,
                 kp=None, kp_density=1E4, expansion=(1, 1, 1), encut=1.3, target_forces=1E-3, output_file=None,
                 energy_tol=1E-3):

        self.ini_factor = ini_factor
        self.fin_factor = fin_factor
        self.delta_factor = delta_factor
        self.kp_density = kp_density
        self.encut = encut
        self.target_forces = target_forces
        self.output_file = output_file
        self.expansion = expansion
        self.energy_tol = energy_tol

        if kp is None:
            kp = pychemia.crystal.KPoints.optimized_grid(structure.lattice, kp_density=kp_density, force_odd=True)
            self.kpoints = kp
        else:
            self.kpoints = kp
            self.kp_density = kp.get_density_of_kpoints(structure.lattice)

        task_params = {'ini_factor': ini_factor, 'fin_factor': fin_factor, 'delta_factor': delta_factor,
                       'kp_density': kp_density, 'expansion': expansion, 'encut': encut,
                       'target_forces': target_forces, 'kp': kp.to_dict}
        Task.__init__(self, structure=structure, task_params=task_params, executable=executable, workdir=workdir)

        self.output = []

    def run(self, nparal=4):

        for ifactor in np.arange(self.ini_factor, self.fin_factor + 0.9 * self.delta_factor, self.delta_factor):

            lattice = self.structure.lattice
            new_lengths = (ifactor - 1.0) * np.array(self.expansion) * lattice.lengths + lattice.lengths
            newlattice_params = tuple(np.concatenate(new_lengths, lattice.angles))
            newlattice = pychemia.crystal.Lattice.from_parameters_to_cell(*newlattice_params)
            newst = pychemia.Structure(cell=newlattice.cell, symbols=self.structure.symbols,
                                       reduced=self.structure.reduced)

            tmpkp = pychemia.crystal.KPoints.optimized_grid(newst.lattice, kp_density=self.kp_density, force_odd=True)

            if ifactor < 1.0:
                print('\nCompresing cell to %7.3f percent' % (ifactor * 100))
                print('-------------------\n')
            elif ifactor > 1.0:
                print('\nExpanding cell to %7.3f percent' % (ifactor * 100))
                print('-------------------\n')
            else:
                print('\nOriginal cell')
                print('-------------\n')

            new_density = tmpkp.get_density_of_kpoints(newst.lattice)
            print('KP Density: target: %d  new lattice: %d' % (self.kp_density, new_density))
            print('KP Grid:    target: %s  new lattice: %s' % (self.kpoints.grid, tmpkp.grid))

            if np.prod(tmpkp.grid) > np.prod(self.kpoints.grid):

                print('The new cell ask for a bigger grid, doing a new convergence...\n')

                self.cleaner()
                print('\nConvergence of K-Point Grid')
                print('---------------------------\n')
                ck = ConvergenceKPointGrid(newst, workdir=self.workdir + os.sep + 'KPCONV_' + str(ifactor),
                                           executable=self.executable, encut=self.encut, energy_tolerance=self.energy_tol,
                                           recover=True)
                ck.run(nparal=nparal)
                ck.save()
                ck.report()
                self.kpoints = ck.best_kpoints
            else:
                print('The current grid is still good')

            self.cleaner()
            relax = IonRelaxation(structure=newst, workdir=self.workdir + os.sep + 'RELAX_' + str(ifactor),
                                  kp_grid=self.kpoints.grid, encut=self.encut,
                                  relax_cell=False, target_forces=self.target_forces, waiting=False,
                                  executable=self.executable)
            relax.run(nparal=nparal)

            vo = pychemia.code.vasp.VaspOutput(self.workdir + '_' + str(ifactor) + '/OUTCAR')
            relst = relax.get_final_geometry()
            symm = pychemia.symm.StructureSymmetry(relst)

            self.output.append({'factor': ifactor, 'volume': newst.volume, 'density': newst.density,
                                'grid': list(self.kpoints.grid), 'output': vo.to_dict, 'spacegroup': symm.number()})
            self.save()

    def plot(self, filedir=None, file_format='pdf'):
        if filedir is None:
            filedir = self.workdir
        import matplotlib.pyplot as plt
        plt.switch_backend('agg')

        plt.figure(figsize=(8, 6))
        plt.subplots_adjust(left=0.1, bottom=0.08, right=0.95, top=0.95, wspace=None, hspace=None)
        data = np.array([[x['output']['energy'] / self.structure.natom, x['volume'] / self.output[0]['volume']]
                         for x in self.output])

        plt.plot(data[:, 0], data[:, 1])

        plt.savefig(filedir + os.sep + 'stress.' + file_format)

    def load(self, filename=None):
        if filename is None:
            filename = self.workdir + os.sep + 'task.json'
        rf = open(filename)
        data = json.load(rf)
        rf.close()
        self.task_params = data['task_params']
        self.output = data['output']
        self.ini_factor = self.task_params['ini_factor']
        self.fin_factor = self.task_params['fin_factor']
        self.delta_factor = self.task_params['delta_factor']
        self.kp_density = self.task_params['kp_density']
        self.expansion = self.task_params['expansion']
        self.encut = self.task_params['encut']
        self.target_forces = self.task_params['target_forces']
        self.kpoints = pychemia.crystal.KPoints.from_dict(self.task_params['kp'])

    def report(self, file_format='html'):
        from lxml.builder import ElementMaker, E
        self.plot(filedir=self.report_dir, file_format='jpg')

        element_maker = ElementMaker(namespace=None, nsmap={None: "http://www.w3.org/1999/xhtml"})
        html = element_maker.html(E.head(E.title("VASP Ideal Strength")),
                                  E.body(E.h1("VASP Ideal Strength"),
                                         E.h2('Structure'),
                                         E.pre(str(self.structure)),
                                         E.h2('Ideal Strength'),
                                         E.p(E.img(src='strenth.jpg', width="800", height="600", alt="Strength")),
                                         ))

        return self.report_end(html, file_format)

    def cleaner(self):

        files = [x for x in os.listdir(self.workdir) if os.path.isfile(self.workdir + os.sep + x)]
        for i in files:
            if i[:5] in ['INCAR', 'POSCA', 'OUTCA', 'vaspr']:
                os.remove(self.workdir + os.sep + i)



"""
Routines related to Report generation
"""
from pychemia import HAS_MATPLOTLIB, HAS_MAYAVI
from .dos import DensityOfStates, plot_one_dos, plot_many_dos
from .povray import StructurePovray

if HAS_MAYAVI:
    from .structure_plot import StructurePlot

# from _pyprocar import BandStructure

if HAS_MATPLOTLIB:
    from pychemia.visual.searcher import plot_evolution_circular

# __all__ = filter(lambda s: not s.startswith('_'), dir())



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import argparse
import subprocess
import time
import shutil
import socket
import numpy as np
from pychemia.code.abinit import AbinitInput, AbinitOutput
from pychemia.population.orbitaldftu import dmatpawu2params


def which(program):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Orbital DFTU Executor, execute several ABINIT runs changing the "
                                                 "value from dmatpawu from the previous output as input until a "
                                                 "tolerance is reached. The executor assumes the existance of a "
                                                 "command 'abinit' and 'mpirun' and files 'abinit.in' and "
                                                 "'abinit.files' the canonical names for those files in ABINIT")

    parser.add_argument('--usedmatpu', type=int, help='ABINIT Variable usedmatpu for each run (default: 25)', 
                        required=False, default=25, metavar='<N>')
    parser.add_argument('--nstep', type=int, help='ABINIT Variable nstep for each run (default: 50)', 
                        required=False, default=50, metavar='<N>')
    parser.add_argument('--tolvrs', type=float, help='ABINIT Variable tolvrs for each run (default: 1E-14)', 
                        required=False, default=1E-14, metavar='<X>')
    parser.add_argument('--target_nres2', type=float, help='Stopping criteria for this executor (default: 1E-12)', 
                        required=False, default=1E-12, metavar='<X>')
    parser.add_argument('--max_nruns', type=int, help='Maximum number of runs allowed (default: 10)', 
                        required=False, default=10, metavar='<N>')
    parser.add_argument('--nhours', type=int, help='Maximun number of hours, ignored if running through a queue '
                                                   'system (PBS), mandatory otherwise',
                        required=False, default=0, metavar='<N>')
    parser.add_argument('--nparal', type=int, help='Number of cores for use with MPI, ignored if running through a '
                                                   'queue system (PBS), mandatory otherwise',
                        required=False, default=0, metavar='<N>')

    args = parser.parse_args()

    print(" ABINIT Orbital DFT+U Executor")
    print(" =============================\n")

    print("Running on: %s" % socket.gethostname())
    print("Current load: %f %f %f" % os.getloadavg())

    # Checking the consistency of all arguments:
    if args.target_nres2 <= args.tolvrs:
        print("Target value must be bigger than ABINIT internal criteria for tolvrs")
        parser.print_help()
        exit(1)
    if args.usedmatpu >= args.nstep:
        print("Total number of SCF steps 'nstep' must be bigger than 'usedmatpu' the number of steps with 'dmatpawu' "
              "fixed")
        parser.print_help()
        exit(1)

    if not os.path.exists('abinit.in'):
        with open('ERROR', 'w') as wf:
            wf.write('No abinit.in')
        raise RuntimeError("File 'abinit.in' could not be found or its symbolic link is broken")
    if not os.path.exists('abinit.files'):
        with open('ERROR', 'w') as wf:
            wf.write('No abinit.files')
        raise RuntimeError("File 'abinit.files' could not be found or its symbolic link is broken")

    # Checking the existance of "mpirun" and "abinit"
    ret = which('mpirun')
    if ret is None:
        with open('ERROR', 'w') as wf:
            wf.write('No mpirun executable')
        raise RuntimeError("Command 'mpirun' could not be found, maybe you need to load the module first")
    print("mpirun: %s" % ret)    
    ret = which('abinit')
    if ret is None:
        with open('ERROR', 'w') as wf:
            wf.write('No abinit executable')
        raise RuntimeError("Command 'abinit' could not be found, maybe you need to load the module first")
    print("abinit: %s" % ret)

    usedmatpu = args.usedmatpu
    nstep = args.nstep
    tolvrs = args.tolvrs
    target_nres2 = args.target_nres2
    max_nruns = args.max_nruns

    nodefile = os.getenv('PBS_NODEFILE')
    nparal = 1
    if nodefile is not None:
        print("Nodefile: %s" % nodefile)
        rf = open(nodefile)
        nparal = len(rf.readlines())
    elif args.nparal > 0:
        nparal = args.nparal
    else:
        print("ERROR: No queue system detected and no positive value for 'nparal'")
        exit(1)

    walltime = os.getenv('PBS_WALLTIME')
    if walltime is not None:
        walltime = int(walltime)
    elif args.nhours > 0:
        walltime = int(args.nhours*3600)
    else:
        print("ERROR: No queue system detected and no positive value for 'nhours'")
        exit(1)

    print("Walltime: %d seconds = %d minutes = %d hours)" % (walltime, int(walltime/60), int(walltime/3600)))
    print("Number of cores for MPI: %d" % nparal)

    # Getting the current time, use to compute the remaining time in execution
    start_time = time.time()

    abi = AbinitInput('abinit.in')
    print("Checking that abinit.in contains value for dmatpawu...", end='')
    if 'dmatpawu' not in abi.variables:
        print('No')
        raise ValueError("ERROR: Could not open abinit.in")
    else:
        print('Yes')

    print("Checking that abinit.in contains value for lpawu...", end='')
    if 'lpawu' not in abi.variables:
        print('No')
        raise ValueError("ERROR: Could not open abinit.in")
    else:
        print('Yes, max lpawu=%d' % max(abi['lpawu']))

    print('Setting ABINIT variables usedmatpu=%d, nstep=%d and tolvrs=%e' % (usedmatpu, nstep, tolvrs))
    abi['usedmatpu'] = usedmatpu
    abi['nstep'] = nstep
    abi['tolvrs'] = tolvrs
    print('Writting modified abinit.in')
    abi.write('abinit.in')

    # Getting the index from the last execution and adding one for the next run
    index = 0
    while True:
        if os.path.isfile('abinit_%02d.in' % index):
            index += 1
        else:
            break
    
    if index >= max_nruns:
        print("Total number of runs has been achieve already, increse 'max_nruns' if you want to continue")
        parser.print_help()
        exit(1)

    print("Executing run with index: %d" % index)
    while index < max_nruns:
        print('\nABINIT execution %d of %d' % (index+1, max_nruns))
        abi = AbinitInput('abinit.in')

        # If possible set the WFK from the output back to input 
        if os.path.isfile('abinit-i_WFK'):
            abi['irdwfk'] = 1
            abi.write('abinit.in')

        # Calling ABINIT
        command_line = "mpirun -np %d abinit < abinit.files > abinit.log 2> abinit.err" % nparal
        print('Running; %s' % command_line)
        start_run = time.time()
        subprocess.call(command_line, shell=True)
        end_run = time.time()

        # Delete the error file if empty
        if os.path.isfile('abinit.err') and os.path.getsize('abinit.err') == 0:
            os.remove('abinit.err')

        runtime = end_run-start_run
        print('Execution finished, execution took %d minutes' % int(runtime/60))

        # If everything works fine with ABINIT we have abinit.out
        # Otherwise is better to stop the entire run
        if not os.path.isfile('abinit.out'):
            with open('ERROR', 'w') as wf:
                wf.write('No abinit.out')
            raise ValueError('File not found: abinit.out')

        if not os.path.isfile('abinit.log'):
            with open('ERROR', 'w') as wf:
                wf.write('No abinit.log')
            raise ValueError('File not found: abinit.log')

        if os.path.isfile('abinit.in'):
            shutil.copy2('abinit.in', 'abinit_%02d.in' % index)

        # Renaming logs and setting WFK back to input
        if os.path.isfile('abinit.log'):
            print("Renaming abinit.log")
            os.rename('abinit.log', 'abinit_%02d.log' % index)
        if os.path.isfile('abinit.err'):
            print("Renaming abinit.err")
            os.rename('abinit.err', 'abinit_%02d.err' % index)
        if os.path.isfile('abinit-o_WFK'):
            print("Renaming abinit-o_WFK")
            os.rename('abinit-o_WFK', 'abinit-i_WFK')
        if os.path.isfile('abinit.out'):
            print("Renaming abinit.out")
            shutil.copy2('abinit.out', 'abinit_%02d.txt' % index)
            os.rename('abinit.out', 'abinit_%02d.out' % index)
            abiout = 'abinit_%02d.out' % index
        else:
            print("Could not find abinit.out")

        # Opening the output file
        print("Reading the output from 'abinit.out'...")
        abo = AbinitOutput(abiout)
        if not abo.is_finished:
            print("abinit.out is truncated, discarting that output redoing the calculation")
            continue

        # The final density matrix is build from the outputi
        ndim = 2*max(abi['lpawu'])+1

        params = dmatpawu2params(abi['dmatpawu'], ndim)
        print("Euler angles from 'abinit.in':")
        for i in params['euler_angles']:
            for j in i:
                print(" %7.3f" % j, end=' ')
            print("\n", end='')

        dmatpawu = abo.get_dmatpawu()
        if dmatpawu is not None:
            params = dmatpawu2params(dmatpawu, ndim)
            print("Euler angles from 'abinit.out':")
            for i in params['euler_angles']:
                for j in i:
                    print(" %7.3f" % j, end=' ')
                print("\n", end='')
        else:
            print("Could not get final dmatpawu from abinit.out")
            dmatpawu = np.array(abi['dmatpawu']).reshape((-1, ndim, ndim))

        print('Shape of dmatpawu: %s' % str(dmatpawu.shape))

        # Updating dmatpawu from the output back to input
        abi['dmatpawu'] = list(dmatpawu.flatten())
        if os.path.isfile('abinit-i_WFK'):
            abi['irdwfk'] = 1
        abi.write('abinit.in')

        # Checking if you should accept the current residual
        energetics = abo.get_energetics()
        if energetics is None:
            raise RuntimeError("Could not get energetics from the output")
        nres2 = 1.0
        nres2 = energetics['nres2'][-1]

        if nres2 < target_nres2 or index == max_nruns-1:
            wf = open('COMPLETE', 'w')
            wf.write("%d\n" % index)
            wf.close()
            break

        # Current time
        curtime = time.time()
        if curtime+runtime > start_time + walltime:
            print("Based on previous run, it is unlikely that next run will have time to complete, exiting")
            print("The walltime for this job is %d minutes and we have been running for %d minutes" %
                  (int(walltime/60), int((curtime-start_time)/60)))
            break
        else:
            print("Remaining time %d minutes, time for one more run" % int((start_time + walltime - curtime)/60))

        # Incresing index for next run
        index += 1



import os
from scipy.io import netcdf_file
from subprocess import Popen, PIPE


class Multibinit:
    """
    Create a Multibinit object from a given multibinit "files" file. From the files it interpret the location and read
    the contents for the several input and output files for multibinit executions.
    """
    # The default names coms from Documentation:
    # https://docs.abinit.org/guide/multibinit/
    basedir = "."
    files = "multibinit.files"
    files_list = {'in': 'multibinit.in',
                  'out': 'multibinit.out',
                  'model': 'model.XML',
                  'anharmonic': 'model_anharmonic.XML',
                  'training': 'training_set_HIST.nc'}
    input = {}

    def __init__(self, files='multibinit.files'):
        """
        Creates an Multibinit object readin the "files" file
        """
        self.files = os.path.basename(files)
        if os.path.exists(files):
            self.basedir = os.path.dirname(os.path.abspath(files))
            self.read_files()
        if os.path.isfile(self.basedir + os.sep + self.files_list['in']):
            self.read_input()

    def read_files(self):
        if not os.path.isfile(self.basedir + os.sep + self.files):
            raise ValueError("Multibinit 'files' file could not be read.")
        with open(self.basedir + os.sep + self.files) as rf:
            data = rf.readlines()
            self.files_list['in'] = data[0].strip()
            self.files_list['out'] = data[1].strip()
            self.files_list['model'] = data[2].strip()
            self.files_list['anharmonic'] = data[3].strip()
            self.files_list['training'] = data[4].strip()
            if self.files_list['anharmonic'].lower() == 'no':
                self.files_list['anharmonic'] = None
            if self.files_list['training'].lower() == 'no':
                self.files_list['training'] = None

    def read_input(self):
        self.input = {}
        rf = open(self.basedir + os.sep + self.files_list['in'])
        data = rf.readlines()
        for iline in data:
            if "#" in iline:
                iline = iline.split("#")[0].strip()
            if '=' in iline:
                variable = iline.split('=')[0].strip()
                raw_value = iline.split('=')[1].strip()
                if variable in self.input:
                    raise ValueError("Duplicate variable '%s' in input file")
                if ' ' in raw_value:
                    value_list = raw_value.split()
                    try:
                        value = [int(x) for x in value_list]
                    except ValueError:
                        try:
                            value = [float(x) for x in value_list]
                        except ValueError:
                            value = value_list
                else:
                    try:
                        value = int(raw_value)
                    except ValueError:
                        try:
                            value = float(raw_value)
                        except ValueError:
                            value = raw_value

                self.input[variable] = value

    def read_training(self):
        if self.files_list['training'] is not None:
            f = netcdf_file(self.basedir + os.sep + self.files_list['training'])

    def write_input(self, filename):
        with open(filename, 'w') as wf:
            for ivariable in sorted(self.input.keys()):
                value = self.input[ivariable]
                if isinstance(value, list):
                    wf.write("%s = " % ivariable)
                    for i in value:
                        wf.write("%s " % i)
                    wf.write("\n")
                elif isinstance(value, int):
                    wf.write("%s = %d\n" % (ivariable, value))
                elif isinstance(value, float):
                    wf.write("%s = %f\n" % (ivariable, value))
                else:
                    wf.write("%s = %s\n" % (ivariable, value))

    def write_files(self, filename):
        with open(filename, 'w') as wf:
            for ikey in ['in', 'out', 'model', 'anharmonic', 'training']:
                if ikey in ['anharmonic', 'training'] and self.files_list[ikey] is None:
                    wf.write('no\n')
                else:
                    wf.write(self.files_list[ikey] + '\n')

    def run(self, executable='multibinit', nparal=4, output_file=None, show_output=False):

        if nparal < 2:
            command = "%s < %s" % (executable, self.files)
        else:
            command = "mpirun -np %d %s < %s" % (nparal, executable, self.files)

        with Popen(command, shell=True, stdout=PIPE) as proc:
            ret = proc.stdout.read()
        if show_output:
            print(ret.decode())
        if output_file is not None:
            wf = open(output_file, 'w')
            wf.write(ret.decode())
            wf.close()
        return ret.decode()



from numpy import loadtxt
from pychemia.utils.periodic import atomic_symbol
from pychemia import Structure


def load(filename):
    symbols = loadtxt(filename, skiprows=2, usecols=[0], dtype='|S2', ndmin=1)
    symbols = [x.decode('utf-8') for x in symbols]

    for i in range(len(symbols)):
        if symbols[i].isdigit():
            symbols[i] = atomic_symbol(int(symbols[i]))

    positions = loadtxt(filename, skiprows=2, usecols=(1, 2, 3), ndmin=2)
    natom = len(symbols)
    periodicity = 3 * [False]
    return Structure(symbols=symbols, positions=positions, periodicity=periodicity, natom=natom)


def save(structure, filename):

    if isinstance(structure, Structure):
        sts = [structure]
    else:
        sts = structure

    wf = open(filename, 'w')
    for st in sts:
        xyz = str(st.natom) + '\n\n'
        for i in range(structure.natom):
            xyz += " %2s %15.7f %15.7f %15.7f\n" % (structure.symbols[i],
                                                    structure.positions[i, 0],
                                                    structure.positions[i, 1],
                                                    structure.positions[i, 2])
        wf.write(xyz)
    wf.close()




import itertools
import math
from math import gcd
from math import cos, sin, sqrt
import numpy as np
import pprint


class FormatPrinter(pprint.PrettyPrinter):

    def __init__(self, formats):
        super(FormatPrinter, self).__init__()
        self.formats = formats

    def format(self, obj, ctx, maxlvl, lvl):
        if type(obj) in self.formats:
            return self.formats[type(obj)] % obj, 1, 0
        return pprint.PrettyPrinter.format(self, obj, ctx, maxlvl, lvl)


def length_vector(v):
    """
    Returns the length of a vector 'v' in arbitrary number of dimensions

    :param v: (list, numpy.ndarray) Vector to compute length

    :rtype : (float) The lenght of the vector

    Examples

    >>> length_vector([1, 2, 3])
    3.7416573867739413

    """
    return np.linalg.norm(v)


def length_vectors(m):
    """
    Returns the lengths of several vectors
    arranged as rows in a MxN matrix

    :param m: numpy.ndarray

    :rtype : numpy.ndarray

    Examples
    >>> length_vectors([[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 0, 0], [0, 0, 2]]) # doctest: +SKIP
    array([  3.74165739,   8.77496439,  13.92838828,   1.        ,   2.        ])

    """
    m = np.array(m)
    return np.linalg.norm(m, axis=1)


def unit_vector(v):
    """
    Returns the unit vector of the vector.
    Arbitrary number of dimensions

    :param v: list, numpy.array

    :rtype : numpy.ndarray

    Examples
    >>> a = unit_vector([1, 2, 3])

    >>> a # doctest: +SKIP
    array([ 0.26726124,  0.53452248,  0.80178373])

    >>> length_vector(a)
    1.0

    """
    if length_vector(np.array(v, dtype=float)) < 1E-10:
        raise ValueError('Vector is null')
    return np.array(v) / length_vector(np.array(v, dtype=float))


def unit_vectors(m):
    """
    Returns the unit vectors of a set
    of vectors arranged as rows in MxN matrix

    :param m: numpy.ndarray

    :rtype : numpy.ndarray

    Example
    >>> from pychemia.utils.mathematics import *

    >>> b = unit_vectors([[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 0, 0], [0, 0, 2]])

    >>> b # doctest: +SKIP
    array([[ 0.26726124,  0.53452248,  0.80178373],
           [ 0.45584231,  0.56980288,  0.68376346],
           [ 0.50257071,  0.57436653,  0.64616234],
           [ 1.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  1.        ]])

    >>> max(length_vectors(b))-1.0<0.1
    True

    >>> 1.0-min(length_vectors(b))<0.1
    True

    """
    m = np.array(m)
    return m / (np.linalg.norm(m, axis=1)[:, np.newaxis])


def angle_vector(v1, v2, units='rad'):
    """
    Returns the angle in radians (default) or degrees
    between vectors 'v1' and 'v2'::

    :param v1: (list, numpy.ndarray)
    :param v2: (list, numpy.ndarray)
    :param units: (str) : 'rad' (default) Radians, 'deg' Degrees

    :rtype : float

    Examples:
    >>> angle_vector([1, 0, 0], [0, 1, 0])
    1.5707963267948966
    >>> angle_vector([1, 0, 0], [1, 0, 0])
    0.0
    >>> round(angle_vector([1, 0, 0], [-1, 0, 0]),15) # doctest: +SKIP
    3.141592653589793
    >>> angle_vector([1, 0, 0], [0, 1, 0], units='deg')
    90.0
    >>> angle_vector([1, 0, 0], [-1, 0, 0], units='deg')
    180.0

    """
    assert (units in ['rad', 'deg'])

    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    dot = np.dot(v1_u, v2_u)
    if dot > 1.0:
        dot = 1.0
    angle = np.arccos(dot)
    if np.isnan(angle):
        if (v1_u == v2_u).all():
            return 0.0
        else:
            return np.pi
    if units == 'rad':
        return angle
    elif units == 'deg':
        return 180.0 * angle / np.pi


def angle_between_vectors(a, b):
    assert a.shape == b.shape

    norms = (np.linalg.norm(a, axis=1) * np.linalg.norm(b, axis=1))
    norms[norms == 0] = 1
    dots = np.sum(a * b, axis=1) / norms
    dots = np.round(dots, 15)
    dots[dots > 1.0] = 1.0
    norms = (np.linalg.norm(a, axis=1) * np.linalg.norm(b, axis=1))
    dots[norms == 0] = 1.0
    return np.arccos(dots)


def angle_vectors(m, units='rad'):
    """
    Returns all the angles for all the
    vectors arranged as rows in matrix 'm'

    :param m: (numpy.ndarray)
    :param units: (str) : 'rad' Radians, 'deg' Degrees

    :rtype : numpy.ndarray

    Examples:
    >>> a = angle_vectors([[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 0, 0], [0, 0, 2]])

    >>> pprint.pprint(a) # doctest: +SKIP
    {(0, 1): 0.2257261285527342,
     (0, 2): 0.2858867976945072,
     (0, 3): 1.3002465638163236,
     (0, 4): 0.6405223126794245,
     (1, 2): 0.060160669141772885,
     (1, 3): 1.0974779950809703,
     (1, 4): 0.8178885561654512,
     (2, 3): 1.0442265974045177,
     (2, 4): 0.8682510378027637,
     (3, 4): 1.5707963267948966}


    >>> a = angle_vectors([[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 0, 0], [0, 0, 2]], units='deg')

    >>> pprint.pprint(a) # doctest: +SKIP
    {(0, 1): 12.933154491899135,
     (0, 2): 16.380106926405656,
     (0, 3): 74.498640433063002,
     (0, 4): 36.699225200489877,
     (1, 2): 3.4469524345065143,
     (1, 3): 62.880857226618922,
     (1, 4): 46.861562380328941,
     (2, 3): 59.829776886585428,
     (2, 4): 49.747120023952057,
     (3, 4): 90.0}

    """

    ret = {}
    for i in itertools.combinations(range(len(m)), 2):
        ret[i] = angle_vector(m[i[0]], m[i[1]], units=units)
    return ret


def distance(v1, v2):
    """
    Return the vector v2-v1, the vector going from v1 to v2
    and the magnitude of that vector.

    :param v1: (list, numpy.ndarray)
    :param v2: (list, numpy.ndarray)

    :rtype : tuple

    Examples:
    >>> distance([0, 0, 0, 1], [1, 0, 0, 0])
    (array([ 1,  0,  0, -1]), 1.4142135623730951)
    >>> distance([-1, 0, 0], [1, 0, 0])
    (array([2, 0, 0]), 2.0)

    """
    ret = np.array(v2) - np.array(v1)
    return ret, length_vector(ret)


def distances(m):
    """
    Return all the distances for all possible combinations of the row vectors in matrix m

    :param m: (list, numpy.ndarray)

    :rtype : dict

    Example:
    >>> import pprint

    >>> pprint.pprint(distances([[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 0, 0], [0, 0, 2]])) # doctest: +SKIP
    {(0, 1): (array([3, 3, 3]), 5.196152422706632),
     (0, 2): (array([6, 6, 6]), 10.392304845413264),
     (0, 3): (array([ 0, -2, -3]), 3.6055512754639891),
     (0, 4): (array([-1, -2, -1]), 2.4494897427831779),
     (1, 2): (array([3, 3, 3]), 5.196152422706632),
     (1, 3): (array([-3, -5, -6]), 8.3666002653407556),
     (1, 4): (array([-4, -5, -4]), 7.5498344352707498),
     (2, 3): (array([-6, -8, -9]), 13.45362404707371),
     (2, 4): (array([-7, -8, -7]), 12.727922061357855),
     (3, 4): (array([-1,  0,  2]), 2.2360679774997898)}

    """
    ret = {}
    for i in itertools.combinations(range(len(m)), 2):
        ret[i] = distance(m[i[0]], m[i[1]])
    return ret


def wrap2_pmhalf(x):
    """
    Wraps a number or array in the interval ]-1/2, 1/2] values = -1/2 will be wrapped  to 1/2

    :param x: (float) The number to be wrapped in the interval (-1/2, 1/2]

    Examples:

    >>> wrap2_pmhalf(-0.5)
    0.5
    >>> wrap2_pmhalf(0.0)
    0.0
    >>> wrap2_pmhalf([-0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75])
    array([ 0.25,  0.5 , -0.25,  0.  ,  0.25,  0.5 , -0.25])
    >>> wrap2_pmhalf([[-0.75, -0.5, -0.25], [0.25, 0.5, 0.75]])
    array([[ 0.25,  0.5 , -0.25],
           [ 0.25,  0.5 , -0.25]])

    """

    def wrap(num):
        tol12 = 1e-12
        if num > 0:
            ret = (num + 0.5 - tol12) % 1.0 - 0.5 + tol12
        else:
            ret = -(-(num - 0.5 - tol12) % 1.0) + 0.5 + tol12
        for y in [-0.25, 0.0, 0.25, 0.5]:
            ret = (lambda num2: y if abs(y - num2) < tol12 else num2)(ret)
        return ret

    if np.iterable(x):
        vec = np.vectorize(wrap)
        return vec(x)
    else:
        return wrap(x)


def vector_set_perpendicular(vector3):
    """
    Produces a set of three mutually perpendicular vectors
    The two other vectors will be unitary

    :return: (tuple) Two numpy arrays
    """
    v1 = unit_vector(vector3)
    v2 = None
    v3 = None
    while True:
        other = unit_vector(np.random.random_sample(3))
        if np.abs(np.dot(v1, other)) > 0.05:
            v2 = unit_vector(np.cross(v1, other))
            v3 = unit_vector(np.cross(v1, v2))
            break
        else:
            continue
    # print _np.dot(v1, v2)
    # print _np.dot(v1, v3)
    # print _np.dot(v2, v3)

    # assert (_np.abs(_np.dot(v1, v2)) < 1E-15)
    # assert (_np.abs(_np.dot(v1, v3)) < 1E-15)
    # assert (_np.abs(_np.dot(v2, v3)) < 1E-15)
    return v1, v2, v3


def matrix_from_eig(v1, v2, v3, lam1, lam2, lam3):
    """
    Given 3 eigenvectors and 3 eigenvalues, returns the
    matrix A that has those eigenvectors and eigenvalues.

    The matrix $A = P.D.P^{-1}$

    Where P is the column stack of eigenvectors and
    D is a diagonal matrix of eigevalues

    :param v1: First eigenvector
    :param v2: Second eigenvector
    :param v3: Third eigenvector
    :param lam1: First eigenvalue
    :param lam2: Second eigenvalue
    :param lam3: Third eigenvalue
    :return: (numpy.ndarray) The matrix
    """
    matrixp = np.vstack((v1, v2, v3)).T
    matrixd = np.diag([lam1, lam2, lam3])
    matrixpinv = np.linalg.inv(matrixp)
    matrixa = np.dot(matrixp, np.dot(matrixd, matrixpinv))
    return matrixa


def integral_gaussian(a, b, mu, sigma):
    """
    Computes the integral of a gaussian centered
    in mu with a given sigma
    :param a:
    :param b:
    :param mu:
    :param sigma:
    :return:
    """

    # Integral from -\infty to a
    val_floor = 0.5 * (1 + math.erf((a - mu) / (sigma * math.sqrt(2.0))))

    # Integral from -\infty to b
    val_ceil = 0.5 * (1 + math.erf((b - mu) / (sigma * sqrt(2.0))))

    return val_ceil - val_floor


def frexp10(x):
    exp = int(math.floor(math.log10(abs(x))))
    return x / 10 ** exp, exp


def round_small(number, ndigits=0):
    mantissa, exponent = frexp10(number)
    mantissa = round(mantissa, ndigits)
    return mantissa * 10 ** exponent


def sieve_atkin(limit):
    """
    Computes all prime numbers up to a 'limit'
    using the Sieve of Atkin

    :param limit:
    :return:

    Example:
    >>> p=sieve_atkin(300)

    >>> len(p)
    63
    >>> p[-1]
    293

    """

    ret = [2, 3]
    sieve = [False] * (limit + 1)
    for x in range(1, int(sqrt(limit)) + 1):
        for y in range(1, int(sqrt(limit)) + 1):
            n = 4 * x * x + y * y
            if n <= limit and (n % 12 == 1 or n % 12 == 5):
                sieve[n] = not sieve[n]
            n = 3 * x * x + y * y
            if n <= limit and n % 12 == 7:
                sieve[n] = not sieve[n]
            n = 3 * x * x - y * y
            if x > y and n <= limit and n % 12 == 11:
                sieve[n] = not sieve[n]
    for x in range(5, int(sqrt(limit))):
        if sieve[x]:
            for y in range(x * x, limit + 1, x * x):
                sieve[y] = False
    for p in range(5, limit):
        if sieve[p]:
            ret.append(p)
    return ret


def trial_division(n):
    """
    Return a list of the prime factors for a natural number
    uses the Sieve of Atkin as a list of primes

    :param n: (int) A natural number

    :rtype : (list)
    """
    if n == 1:
        return [1]
    primes = sieve_atkin(int(n ** 0.5) + 1)
    prime_factors = []

    for p in primes:
        if p * p > n:
            break
        while n % p == 0:
            prime_factors.append(p)
            n /= p
    if n > 1:
        prime_factors.append(n)

    return prime_factors


def lcm(a, b):
    """
    Return the Least Common Multiple
    the smallest positive integer that is divisible by both a and b
    :param a: (int)
    :param b: (int)
    :return: (int)

    :rtype : (int)
    """
    return a * b / gcd(a, b)


def shortest_triple_set(n):
    """
    Return the smallest three numbers (a,b,c) such as
    a*b*c=n
    And a+b+c is as small as possible

    :param n:
    :return:
    """
    # First Compute the prime factors
    prime_factors = trial_division(n)

    if len(prime_factors) == 3:
        # No choice return the three numbers
        return prime_factors
    elif len(prime_factors) < 3:
        # If there are less than 3 complete the set with ones
        while len(prime_factors) % 3 != 0:
            prime_factors = [1] + prime_factors
        return prime_factors
    else:
        factors = np.array(prime_factors)
        while len(factors) > 3:
            # print factors
            # Complete a multiple of 6 and sum folding lowest with highest
            while len(factors) % 6 != 0:
                factors = np.concatenate(([1], factors))
            # Take the first half
            low = factors[:int(len(factors) / 2)]
            # take the second half and invert the order
            high = factors[int(len(factors) / 2):][::-1]
            # Sum both arrays and sort them before reenter
            factors = np.sort(low * high)
        return [int(x) for x in factors]


def rotation_matrix_around_axis_angle(axis, theta):
    """
    Return the rotation matrix needed to rotate any vector
    around the 'axis' an angle of 'theta' radians
    """
    # Given a unit vector u = (ux, uy, uz), where ux**2 + uy**2 + uz**2 = 1,
    u = unit_vector(axis)
    # with ux is the cross product matrix
    ux = np.array([[0, -u[2], u[1]], [u[2], 0, -u[0]], [-u[1], u[0], 0]])
    # uxu is the tensor product of u
    uxu = np.tensordot(u, u.T, axes=0)
    # This is a matrix form of Rodrigues 'rotation formula'
    return np.cos(theta) * np.identity(3) + np.sin(theta) * ux + (1 - np.cos(theta)) * uxu


def rotate_towards_axis(vector, axis, theta=None, fraction=None):
    # If the vector is already parallel to the axis, do nothing
    if np.abs(angle_vector(vector, axis)) < 1E-10 or np.abs(angle_vector(vector, axis) - np.pi) < 1E-10:
        return vector

    # Create a unitary vector perpendicular to the plane created by vector and axis
    uv = unit_vector(np.cross(vector, axis))

    # Use the rodrigues formula to rotate around the vector perpendicular
    if theta is not None:
        m = rotation_matrix_around_axis_angle(uv, theta)
        return np.dot(m, vector)

    if fraction is not None:
        theta = angle_vector(vector, axis)
        m = rotation_matrix_around_axis_angle(uv, fraction * theta)
        return np.dot(m, vector)


def rotation_x(theta):
    """
    Create a rotation matrix around the 'x' axis

    :param theta: (float) Angle in radians
    :return: (numpy.ndarray)

    Examples:
    >>> import numpy as np

    >>> m = rotation_x(np.pi/3)

    >>> np.max(np.dot(m.T, m) - np.eye(3)) < 1E-10
    True

    """
    return np.array([[1, 0, 0],
                     [0, np.cos(theta), -np.sin(theta)],
                     [0, np.sin(theta), np.cos(theta)]])


def rotation_y(theta):
    """
    Create a rotation matrix around the 'y' axis

    :param theta: (float) Angle in radians
    :return: (numpy.ndarray)

    Example:
    >>> import numpy as np

    >>> m = rotation_y(np.pi/3)

    >>> np.max(np.dot(m.T, m) - np.eye(3)) < 1E-10
    True

    """
    return np.array([[np.cos(theta), 0, np.sin(theta)],
                     [0, 1, 0],
                     [-np.sin(theta), 0, np.cos(theta)]])


def rotation_z(theta):
    """
    Create a rotation matrix around the 'z' axis

    :param theta: (float) Angle in radians
    :return: (numpy.ndarray)

    Examples:
    >>> import numpy as np
    >>> m = rotation_z(np.pi/3)

    >>> np.max(np.dot(m.T, m) - np.eye(3)) < 1E-10
    True

    """
    return np.array([[np.cos(theta), -np.sin(theta), 0],
                     [np.sin(theta), np.cos(theta), 0],
                     [0, 0, 1]])


def apply_rotation(vector, theta_x, theta_y, theta_z):
    """
    Apply a rotation matrix to a vector by succesive rotations around
    the three axis 'x', 'y' and 'z'

    :param vector:
    :param theta_x: (float) Angle in radians
    :param theta_y: (float) Angle in radians
    :param theta_z: (float) Angle in radians
    :return: (numpy.ndarray)

  Example:
    >>> a = apply_rotation([0.1, 0.2, 0.3], 3.1415/3, 3.1415/4, 3.1415/5)
    >>> b = apply_rotation(a, -3.1415/3, 0, 0)
    >>> c = apply_rotation(b, 0, -3.1415/4, 0)
    >>> d = apply_rotation(c, 0, 0, -3.1415/5)

    >>> d # doctest: +SKIP
    array([ 0.1,  0.2,  0.3])

    """
    return np.round(np.dot(rotation_x(theta_x), np.dot(rotation_y(theta_y), np.dot(rotation_z(theta_z), vector))), 14)


def rotation_matrix_weave(axis, theta, mat=None):
    try:
        from scipy import weave
    except ImportError:
        raise NotImplementedError
    if mat is None:
        mat = np.eye(3, 3)

    support = "#include <math.h>"
    code = """
        double x = sqrt(axis[0] * axis[0] + axis[1] * axis[1] + axis[2] * axis[2]);
        double a = cos(theta / 2.0);
        double b = -(axis[0] / x) * sin(theta / 2.0);
        double c = -(axis[1] / x) * sin(theta / 2.0);
        double d = -(axis[2] / x) * sin(theta / 2.0);

        mat[0] = a*a + b*b - c*c - d*d;
        mat[1] = 2 * (b*c - a*d);
        mat[2] = 2 * (b*d + a*c);

        mat[3*1 + 0] = 2*(b*c+a*d);
        mat[3*1 + 1] = a*a+c*c-b*b-d*d;
        mat[3*1 + 2] = 2*(c*d-a*b);

        mat[3*2 + 0] = 2*(b*d-a*c);
        mat[3*2 + 1] = 2*(c*d+a*b);
        mat[3*2 + 2] = a*a+d*d-b*b-c*c;
    """

    weave.inline(code, ['axis', 'theta', 'mat'], support_code=support, libraries=['m'])

    return mat


def rotation_matrix_numpy(axis, theta):
    """

    :param axis:
    :param theta:
    :return:

    Example:

    >>> rotation_matrix_numpy([1, 1, 1], 3.1415926/4.0)
    array([[ 0.80473786,  0.50587935, -0.31061722],
           [-0.31061722,  0.80473786,  0.50587935],
           [ 0.50587935, -0.31061722,  0.80473786]])

    """
    uaxis = unit_vector(axis)
    a = cos(theta / 2.)
    b, c, d = -uaxis * sin(theta / 2.)

    return np.array([[a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)],
                     [2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)],
                     [2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c]])


def projector(u, v):
    """
    Computes the projector of 'v' over 'u'
    A vector in the direction of 'u' with a
    magnitude of the projection of 'v' over 'u'

    :param u:
    :param v:
    :return:

    Example:
    >>> projector([0.1, 0.2, 0.3], [0.3, 0.2, 0.1]) # doctest: +SKIP
    array([ 0.07142857,  0.14285714,  0.21428571])
    >>> projector([1, 0, 0], [0, 2, 0]) # doctest: +SKIP
    array([ 0.,  0.,  0.])

    """
    u = np.array(u, dtype=float)
    v = np.array(v, dtype=float)
    return np.dot(v, u) / np.dot(u, u) * np.array(u)


def gram_smith(m):
    """
    Create a Gram-Smith ortoganalized
    matrix, using the first vector of
    matrix 'm' to build the ortogonal
    set of vectors

    :param m:
    :return:

    Example:
    >>> import numpy as np
    >>> o = gram_smith(np.random.rand(3, 3))

    >>> np.round(np.abs(np.linalg.det(o)), 10) == 1.0
    True

    """
    m = np.array(m)
    ret = np.zeros((len(m[0]), len(m[0])))
    ret[0] = m[0] / np.linalg.norm(m[0])
    for k in range(1, len(m[0])):
        ret[k] = m[k]
        for j in range(0, k):
            ret[k] -= projector(ret[j], m[k])
        ret[k] /= np.linalg.norm(ret[k])
    return ret


def gram_smith_qr(ndim=3):
    """
    Create a Gram-Smith orthoganalized matrix.
    The argument is the dimension of the matrix
    and uses a random matrix and QR decomposition
    to build the orthogonal matrix

    :param ndim: Dimension of the matrix
    :return:

    Example:
    >>> import numpy as np
    >>> o = gram_smith_qr(3)
    >>> np.round(np.abs(np.linalg.det(o)), 10)==1.0
    True

    """
    matrix_a = np.random.rand(ndim, ndim)
    while np.linalg.det(matrix_a) < 1E-5:
        matrix_a = np.random.rand(ndim, ndim)
    ret = np.linalg.qr(matrix_a)[0]
    if np.linalg.det(ret) < 0:
        eye = np.eye(ndim)
        eye[0, 0] = -1
        ret = np.dot(ret, eye)
    return ret


def cartesian_to_spherical(xyz):
    """
    Convert a numpy array (Nx3) into
    a numpy array (Nx3) where the first
    column is the magnitude of the vector
    and second and third are spherical angles

    We use the mathematical notation
    (radial, azimuthal, polar)

    :param xyz: numpy.array
    :return: numpy.array
    """
    xyz = np.array(xyz).reshape((-1, 3))
    spherical = np.zeros(xyz.shape)
    xy = xyz[:, 0] ** 2 + xyz[:, 1] ** 2
    spherical[:, 0] = np.sqrt(xy + xyz[:, 2] ** 2)
    spherical[:, 1] = np.arctan2(xyz[:, 1], xyz[:, 0])
    spherical[:, 2] = np.arctan2(np.sqrt(xy), xyz[:, 2])  # for elevation angle defined from Z-axis down
    # spherical[:, 2] = np.arctan2(xyz[:, 2], np.sqrt(xy))   # for elevation angle defined from XY-plane up
    return spherical


def spherical_to_cartesian(spherical):
    """
    Convert a numpy array (Nx3) from
    spherical coordinates to cartesian
    coordinates

    We use the mathematical notation
    (radial, azimuthal, polar)

    :param spherical: numpy.array
    :return: numpy.array
    """
    spherical = np.array(spherical).reshape((-1, 3))
    xyz = np.zeros(spherical.shape)

    xyz[:, 0] = spherical[:, 0] * np.cos(spherical[:, 1]) * np.sin(spherical[:, 2])
    xyz[:, 1] = spherical[:, 0] * np.sin(spherical[:, 1]) * np.sin(spherical[:, 2])
    xyz[:, 2] = spherical[:, 0] * np.cos(spherical[:, 2])
    return xyz


def rotation_ndim(ndim, theta, indices):
    """
    Return a rotation matrix for the Euclidean space in ndim dimensions.
    The angle 'theta' is in radians and the indices should be a list of two
    values where defining the plane of rotation

    :param ndim: Dimension of the matrix
    :param theta: Angle of rotation in radians
    :param indices: Indices of the plane where the rotation takes place.
    :return:
    """

    ret = np.eye(ndim)
    if abs(theta) > 1E-20:
        ret[indices[0], indices[0]] = np.cos(theta)
        ret[indices[1], indices[1]] = np.cos(theta)
        ret[indices[0], indices[1]] = -np.sin(theta)
        ret[indices[1], indices[0]] = np.sin(theta)
    # print('Rotation of angle %f on plane %s\n %s' % (theta, indices, rot))
    return ret


def iterative_rotation_angles(m, nrounds=1):
    # Make a copy of the original matrix
    ident = np.array(m)
    # The dimension of the matrix
    n = m.shape[0]
    # The angles is an ordered list of operations
    # Rotations form a non-abelian group
    rotation_list = []
    inverse = np.eye(n)

    for j in range(nrounds):
        for i in itertools.combinations(range(n), 2):
            # Compute the angle to align the plane i
            # into the canonical base
            num = ident[i[1], i[0]]  # Numerator
            den = ident[i[1], i[1]]  # Denominator
            print('Numerator: %f Denominator: %f' % (num, den))
            if abs(num) < 1E-18:
                theta = 0.0
            elif abs(den) > 1E-18:
                theta = np.arctan(- num / den)
                if den < 0.0:
                    theta += np.pi
            else:
                theta = np.pi/2.0
            rotation_list.append([theta, tuple(i)])
            # Compute a rotation matrix for plane i
            rot = rotation_ndim(n, theta, tuple(i))
            # Apply the rotation left side
            # print('Before rotation  for %s with angle %f \n %s' % (tuple(i), theta, mp))
            ident = np.array(np.dot(ident, rot))
            inverse = np.dot(inverse, rot)
            # print('After rotation for %s with angle %f \n %s' % (tuple(i), theta, mp))

    # rotation_list is a list of Generalized Euler Angles
    # ident is a matrix that should be close to Identity
    # inverse is the return matrix, the matrix that should be
    # close to the inverse (tranpose) of the original matrix
    return rotation_list, ident, inverse


def rotation_matrix_ndim(ndim, rotation_list):
    ret = np.eye(ndim)
    for rotation in rotation_list:
        rot = rotation_ndim(ndim, rotation[0], rotation[1])
        ret = np.dot(ret, rot)
    return ret


def gea_angles(uvector):
    """
    Generalized Euler Angles
    Return the parametric angles decribed on Eq. 1 from the paper

    Generalization of Euler Angles to N-Dimensional Orthogonal Matrices
    David K. Hoffman, Richard C. Raffenetti, and Klaus Ruedenberg
    Journal of Mathematical Physics 13, 528 (1972)
    doi: 10.1063/1.1666011

    :param uvector: A n-dimensional vector
    :return: n dimensional list of angles, the last one is pi/2
    """

    if np.linalg.norm(uvector) != 1.0:
        uvector /= np.linalg.norm(uvector)

    tmp = np.array(uvector)
    angles = []
    n = len(uvector)
    if n < 2:
        raise ValueError('Vector need to be at least dim=2')
    for i in range(n-1):
        if tmp[0] < -1.0:
            tmp[0] = -1.0
        if tmp[0] > 1.0:
            tmp[0] = 1.0
        theta = np.arcsin(tmp[0])
        # print('Sin: %f Angle: %f'% (tmp[0], theta))
        if len(tmp) == 2 and tmp[1] < 0.0:
            if theta > 0:
                theta = np.pi - theta
            elif theta < 0:
                theta = - np.pi - theta
        # print('Sin: %f Angle: %f %s'% (tmp[0], theta, tmp))
        angles.append(theta)
        if np.abs(np.cos(theta)) < 1E-8:
            pass
            # print('ALERT: Zero denominator with numerator: %s' % tmp[1:])
        tmp = tmp[1:]/np.cos(theta)
        if np.abs(np.cos(theta)) < 1E-8:
            pass
            # print('ALERT: Zero denominator with numerator: %s %f' % (tmp, max(tmp)))
    angles.append(np.pi/2.0)
    return angles


"""
Generalization of Euler Angles to N-Dimensional Orthogonal Matrices
David K. Hoffman, Richard C. Raffenetti, and Klaus Ruedenberg
Journal of Mathematical Physics 13, 528 (1972)
doi: 10.1063/1.1666011
"""


def gea_matrix_a(angles):
    """
    Generalized Euler Angles
    Return the parametric angles described on Eqs. 15-19 from the paper:

    Generalization of Euler Angles to N-Dimensional Orthogonal Matrices
    David K. Hoffman, Richard C. Raffenetti, and Klaus Ruedenberg
    Journal of Mathematical Physics 13, 528 (1972)
    doi: 10.1063/1.1666011

    """
    n = len(angles)
    matrix_a = np.eye(n)

    for i in range(n-1):
        matrix_a[i][i] = np.cos(angles[i])
        matrix_a[i][n-1] = np.tan(angles[i])
        for j in range(i+1):
            matrix_a[i][n-1] *= np.cos(angles[j])

    for i in range(n):
        for k in range(n-1):
            if i > k:
                matrix_a[i][k] = -np.tan(angles[i])*np.tan(angles[k])
                for l in range(k, i+1):
                    matrix_a[i][k] *= np.cos(angles[l])

    matrix_a[n - 1][n - 1] = np.tan(angles[n-1])
    for j in range(n):
        matrix_a[n - 1][n - 1] *= np.cos(angles[j])

    return matrix_a


def gea_all_angles(ortho_matrix):
    """
    Generalized Euler Angles
    Return the generalized angles described from the paper

    Generalization of Euler Angles to N-Dimensional Orthogonal Matrices
    David K. Hoffman, Richard C. Raffenetti, and Klaus Ruedenberg
    Journal of Mathematical Physics 13, 528 (1972)
    doi: 10.1063/1.1666011

    :param ortho_matrix: Orthogonal Matrix
    """
    b = np.array(ortho_matrix)
    n = b.shape[0]
    ret = []

    # print('Original matrix\n%s' % b)
    for i in range(1, n):
        # Lets take vectors starting from the last column on ortho_matrix
        an = b[:, -1]
        # print('Vector: \n%s' % an)
        angles = gea_angles(an)
        # print('Angles: \n%s' % angles)
        a = gea_matrix_a(angles)
        # print('Matrix a: \n%s' % a)
        b = np.dot(b.T, a)
        # print('New matrix b: \n%s' % b)
        b = b[:-1, :-1]
        # print('New matrix b (fixed): \n%s' % b)
        ret += angles[:-1]
    return ret


def gea_orthogonal_from_angles(angles_list):
    """
    Generalized Euler Angles
    Return the orthogonal matrix from its generalized angles

    Generalization of Euler Angles to N-Dimensional Orthogonal Matrices
    David K. Hoffman, Richard C. Raffenetti, and Klaus Ruedenberg
    Journal of Mathematical Physics 13, 528 (1972)
    doi: 10.1063/1.1666011

    :param angles_list: List of angles, for a n-dimensional space the total number
                        of angles is k*(k-1)/2
    """

    b = np.eye(2)
    n = int(np.sqrt(len(angles_list)*8+1)/2+0.5)
    tmp = np.array(angles_list)

    # For SO(k) there are k*(k-1)/2 angles that are grouped in k-1 sets
    # { (k-1 angles), (k-2 angles), ... , (1 angle)}
    for i in range(1, n):
        angles = np.concatenate((tmp[-i:], [np.pi/2]))
        tmp = tmp[:-i]
        ma = gea_matrix_a(angles)  # matrix i+1 x i+1
        b = np.dot(b, ma.T).T
        # We skip doing making a larger matrix for the last iteration
        if i < n-1:
            c = np.eye(i+2, i+2)
            c[:-1, :-1] = b
            b = c
    return b



import subprocess
import numpy as np
from pychemia.code.vasp import write_poscar


class PhonopyJob:

    def __init__(self, structure, dim=(2, 2, 2)):

        self.structure = structure
        self.dim = np.array(dim)

    def create_poscars(self):

        write_poscar(self.structure)
        subprocess.call(['phonopy', '-d', r'--dim="%d' % self.dim[0], '%d' % self.dim[1],  r'%d"' % self.dim[2], '-c',
                         'POSCAR'])




from ..codes import CodeRun


class ElkRun(CodeRun):

    def __init__(self, workdir):
        CodeRun.__init__(self, executable='elk', workdir=workdir, use_mpi=True)

    def set_inputs(self):
        pass

    def get_outputs(self):
        pass

    def is_finished(self):
        pass

    def is_successful(self):
        pass



import os
import json
import numpy as np
from ..abinit import AbinitJob
from ...tasks import Task
from ...relaxator import Relaxator
from pychemia.crystal import KPoints
from pychemia import pcm_log

__author__ = 'Guillermo Avendano-Franco'


class IonRelaxation(Relaxator, Task):
    def __init__(self, structure, workdir='.', tolmxf=1E-4, tolrff=1E-2, waiting=False, executable='abinit',
                 ecut=50, kp_grid=None, kp_density=1E4, relax_cell=True, max_calls=10):

        Relaxator.__init__(self, tolmxf)
        self.tolmxf = tolmxf
        self.ecut = ecut
        self.waiting = waiting
        self.abinitjob = AbinitJob(workdir=workdir, executable=executable)
        self.relaxed = False
        self.tolmxf = tolmxf
        self.tolrff = tolrff
        if kp_grid is not None:
            self.kpoints = KPoints(kmode='gamma', grid=kp_grid)
        else:
            self.kpoints = KPoints.optimized_grid(structure.lattice, kp_density=kp_density)
        self.abinitjob.initialize(structure=structure)
        self.relax_cell = relax_cell
        self.max_calls = max_calls
        task_params = {'tolmxf': self.target_forces, 'ecut': self.ecut, 'relax_cell': self.relax_cell,
                       'max_calls': self.max_calls}
        Task.__init__(self, structure=structure, task_params=task_params, workdir=workdir, executable=executable)

    def run(self, nparal=1, verbose=False):
        
        self.abinitjob.set_kpoints(kpoints=self.kpoints)
        self.abinitjob.job_ion_relax(tolmxf=self.tolmxf, tolrff=self.tolrff)
        self.abinitjob.set_ecut(self.ecut)
        self.abinitjob.set_psps()
        self.abinitjob.write_all()
        pcm_log.debug("Starting abinit execution with nparal=%d" % nparal)
        self.abinitjob.run(num_threads=nparal, mpi_num_procs=nparal, verbose=verbose)

    def get_final_geometry(self):
        pass

    def get_forces_stress_energy(self):
        pass

    def plot(self, filedir=None, file_format='pdf'):
        if filedir is None:
            filedir = self.workdir
        import matplotlib.pyplot as plt
        plt.switch_backend('agg')

        plt.figure(figsize=(8, 6))
        plt.subplots_adjust(left=0.1, bottom=0.08, right=0.95, top=0.95, wspace=None, hspace=None)
        forces = np.array(self.output['forces'])
        maxforce = [np.max(np.apply_along_axis(np.linalg.norm, 1, x)) for x in forces]
        avgforce = [np.mean(np.apply_along_axis(np.linalg.norm, 1, x)) for x in forces]

        if np.max(maxforce) > 0.0 and np.max(avgforce) > 0.0:
            plt.semilogy(maxforce, 'b.-', label='Max force')
            plt.semilogy(avgforce, 'r.-', label='Mean force')
        else:
            plt.plot(maxforce, 'b.-', label='Max force')
            plt.plot(avgforce, 'r.-', label='Mean force')
        plt.xlabel('Ion movement iteration')
        plt.ylabel('Max Force')
        plt.savefig(filedir + os.sep + 'forces.' + file_format)
        plt.clf()

        plt.figure(figsize=(8, 6))
        plt.subplots_adjust(left=0.1, bottom=0.08, right=0.95, top=0.95, wspace=None, hspace=None)
        stress = np.array(self.output['stress'])
        diag_stress = [np.trace(np.abs(x)) for x in stress]
        offdiag_stress = [np.sum(np.abs(np.triu(x, 1).flatten())) for x in stress]
        plt.semilogy(diag_stress, 'b.-', label='diagonal')
        plt.semilogy(offdiag_stress, 'r.-', label='off-diagonal')
        plt.legend()
        plt.xlabel('Ion movement iteration')
        plt.ylabel(r'$\sum |stress|$ (diag, off-diag)')
        plt.savefig(filedir + os.sep + 'stress.' + file_format)

    def report(self, file_format='html'):
        from lxml.builder import ElementMaker, E

        self.plot(filedir=self.report_dir, file_format='jpg')

        element_maker = ElementMaker(namespace=None, nsmap={None: "http://www.w3.org/1999/xhtml"})
        html = element_maker.html(E.head(E.title("ABINIT Ion Relaxation")),
                                  E.body(E.h1("ABINIT Ion Relaxation"),
                                         E.h2('Initial Structure'),
                                         E.pre(str(self.structure)),
                                         E.h2('Forces Minimization'),
                                         E.p(E.img(src='forces.jpg', width="800", height="600", alt="Forces")),
                                         E.h2('Stress Minimization'),
                                         E.p(E.img(src='stress.jpg', width="800", height="600", alt="Stress"))
                                         ))

        return self.report_end(html, file_format)

    def load(self, filename=None):
        if filename is None:
            filename = self.workdir + os.sep + 'task.json'
        rf = open(filename)
        data = json.load(rf)
        rf.close()
        self.task_params = data['task_params']
        self.output = data['output']
        self.ecut = self.task_params['ecut']
        self.target_forces = self.task_params['target_forces']
        self.relax_cell = self.task_params['relax_cell']
        self.max_calls = self.task_params['max_calls']



"""
Routines related to convert pymatgen structure into pymatdis structures
"""

from pychemia import HAS_PYMATGEN

if HAS_PYMATGEN:
    from ._pymatgen import cif2structure, pychemia2pymatgen, pymatgen2pychemia

# __all__ = filter(lambda s: not s.startswith('_'), dir())



# coding: utf-8
# Copyright (c) Guillermo Avendano-Franco
# Distributed under the terms of the MIT License.

import os
from re import findall
from numpy import zeros, array, sum
from pychemia import Structure
from pychemia.utils.periodic import atomic_symbols
from itertools import groupby

"""
Routines to read and write POSCAR file
"""

__author__ = "Guillermo Avendano-Franco"
__copyright__ = "Copyright 2017, PyChemia Project"
__version__ = "0.2"
__maintainer__ = "Guillermo Avendano-Franco"
__email__ = "gufranco@mail.wvu.edu"
__date__ = "Feb 20, 2017"


def read_poscar(path='POSCAR'):
    """
    Load a POSCAR file and return a pychemia structure object

    :param path: (str) Path to a POSCAR file or a directory where a file named 'POSCAR' is located
    :return:
    """
    # The argument 'path' could refer to a POSCAR file or the directory where the file 'POSCAR' exists
    if os.path.isfile(path):
        poscarfile = path
        if os.path.dirname(path) != '':
            potcarfile = os.path.dirname(path) + os.sep + 'POTCAR'
        else:
            potcarfile = 'POTCAR'
    elif os.path.isdir(path) and os.path.isfile(path + os.sep + 'POSCAR'):
        poscarfile = path + os.sep + 'POSCAR'
        potcarfile = path + os.sep + 'POTCAR'
    else:
        raise ValueError("ERROR: No POSCAR file found on %s" % path)

    # Reading the POSCAR file
    rf = open(poscarfile, 'r')
    comment = rf.readline().strip()
    latconst = float(rf.readline().split()[0])
    newcell = zeros((3, 3))

    newcell[0, :] = latconst * array([float(x) for x in rf.readline().split()[:3]])
    newcell[1, :] = latconst * array([float(x) for x in rf.readline().split()[:3]])
    newcell[2, :] = latconst * array([float(x) for x in rf.readline().split()[:3]])

    line = rf.readline()
    species = None

    # This is the old format, the only way of knowing which species are refering is by
    # reading the POTCAR file
    natom_per_species = array([int(x) for x in line.split() if x.isdigit()])

    # Check if the file is the new format, in such case this line contains the
    # atomic symbols of the atoms and the next line the number of atoms of each
    # species. The new format makes a POSCAR self-contained to create a structure
    if len(natom_per_species) == 0:
        species = [x for x in line.split() if x in atomic_symbols]
        line = rf.readline()
        natom_per_species = array([int(x) for x in line.split() if x.isdigit()])

    natom = sum(natom_per_species)

    if species is None:
        comment_species = [x for x in comment.split() if x in atomic_symbols]
        if os.path.isfile(potcarfile):
            species = get_species(potcarfile)
        elif len(comment_species) == len(natom_per_species):
            species = comment_species
        else:
            raise ValueError("ERROR: The POSCAR does not have information about the species present on the structure\n"
                             + "You can set a consistent POTCAR along the POSCAR or modify your POSCAR to the\n"
                             + "new format by adding the atomic symbol(s) on the sixth line of the file")

    symbols = []
    for i in range(len(natom_per_species)):
        for j in range(natom_per_species[i]):
            symbols.append(species[i])

    mode = rf.readline()
    if mode[0].lower() in ['c', 'k']:
        kmode = 'Cartesian'
    else:
        kmode = 'Direct'

    pos = []
    for i in range(natom):
        pos += [float(x) for x in rf.readline().split()[:3]]
    pos = array(pos).reshape((-1, 3))

    if kmode == 'Cartesian':
        return Structure(cell=newcell, symbols=symbols, positions=pos, comment=comment)
    else:
        return Structure(cell=newcell, symbols=symbols, reduced=pos, comment=comment)


def write_poscar(structure, filepath='POSCAR', newformat=True, direct=True, comment=None, heterostructure=False):
    """
    Takes an structure from pychemia and save the file
    POSCAR for VASP.

    :param comment: Optional comment to the first line of the POSCAR
    :param structure: (pychemia.Structure) Structure to write POSCAR
    :param filepath: (str) Filename of POSCAR file to create
    :param newformat: (bool) If the new VASP format is used to create the POSCAR
    :param direct: (bool) If True, use reduced coordinates. If False, use cartesian coordinates (default: True)
    """
    comp = structure.get_composition()

    # If heterostructure is true it will keep the repeating order found
    # in the POSCAR. 
    # Added by Uthpala on Apr 20th, 2020.
    if heterostructure:
        species = [i[0] for i in groupby(structure.symbols)]
    else:
        species = get_species_list(structure)
    species_count = [len(list(group)) for key, group in groupby(structure.symbols)]   

    ret = ''
    if comment is None:
        for i in species:
            ret += ' ' + i
    else:
        ret += comment.strip()
    ret += '\n'
    ret += '1.0\n'
    for i in range(3):
        ret += ' %20.16f %20.16f %20.16f\n' % tuple(structure.cell[i])

    if newformat:
        for i in species:
            ret += ' ' + i
        ret += '\n'
    for icount, i in enumerate(species):
        if heterostructure:
           ret += ' ' + str(species_count[icount]) 
        else:
            ret += ' ' + str(comp.composition[i])
    ret += '\n'

    if direct:
        ret += 'Direct\n'
        for i in range(structure.natom):
            ret += ' %20.16f %20.16f %20.16f\n' % tuple(structure.reduced[i])
    else:
        ret += 'Cartesian\n'
        for i in range(structure.natom):
            ret += ' %20.16f %20.16f %20.16f\n' % tuple(structure.positions[i])

    wf = open(filepath, 'w')
    wf.write(ret)
    wf.close()


def get_species_list(structure):
    while True:
        species = []
        for i in structure.symbols:
            if i not in species:
                species.append(i)
        if len(species) == len(structure.species):
            break
        else:
            structure.sort_sites()
    return species


def get_species(path):
    species = []
    rf = open(path, 'r')
    for line in rf.readlines():
        if 'PAW_PBE' in line and 'PAW_PBE' == line.split()[0].strip():
            species.append(line.split()[1].split('_')[0])
        if 'PAW' in line and 'PAW' == line.split()[0].strip() and 'radial' not in line:
            species.append(line.split()[1].split('_')[0])

    return species


def write_potcar(structure, filepath='POTCAR', pspdir='potpaw_PBE', options=None, pspfiles=None, basepsp=None, heterostructure=False):

    # If heterostructure is true it will keep the repeating order found
    # in the POSCAR. 
    # Added by Uthpala on Apr 20th, 2020.
    if heterostructure:
        species = [i[0] for i in groupby(structure.symbols)]
    else:
        species = get_species_list(structure)

    ret = ''
    if basepsp is not None:
        psppath = os.path.abspath(basepsp) + os.sep + pspdir
    else:
        psppath = os.getenv('HOME') + '/.vasp/PP-VASP/' + pspdir
    if not os.path.exists(psppath):
        raise ValueError("The path for VASP Pseudo-potentials does not exists: " + psppath)

    if pspfiles is None:
        pspfiles = []
        for i in species:
            if options is not None and i in options:
                if isinstance(options[i], str):
                    pspfile = psppath + os.sep + i + '_' + options[i] + '/POTCAR'
                elif isinstance(options[i], list):
                    for j in options[i]:
                        pspfile = psppath + os.sep + i + '_' + options[i][j] + '/POTCAR'
                        if os.path.isfile(psppath):
                            break

            else:
                for j in ['', '_sv']:
                    pspfile = psppath + os.sep + i + j + '/POTCAR'
                    if os.path.isfile(pspfile):
                        break
            if not os.path.isfile(pspfile):
                raise ValueError("File not found : " + pspfile)
            pspfiles.append(pspfile)

    for pspfile in pspfiles:
        rf = open(pspfile)
        ret += rf.read()
        rf.close()
    wf = open(filepath, 'w')
    wf.write(ret)
    wf.close()
    return pspfiles


def get_potcar_info(filename='POTCAR'):
    rf = open(filename)
    data = rf.read()
    pairs = findall('([\w ]*)=([.\d ]*)', data)
    ret = {}
    for i in pairs:
        print(i)
        if i[1].strip() == '':
            pass
        elif i[1].strip().isdigit():
            ret[i[0].strip()] = int(i[1])
        elif len(i[1].split()) > 1:
            ret[i[0].strip()] = [float(x) for x in i[1].split()]
        else:
            ret[i[0].strip()] = float(i[1])

    return ret



"""
Includes class LennardJones and several methods to relax and compute forces and energies from Lennard-Jones clusters

"""
# import pyximport
# pyximport.install()

from .lj import LennardJones, lj_compact_evaluate
from .lj_utils import lj_energy, lj_forces, lj_gradient



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import logging
import argparse
import time
import pychemia
from pychemia.runner import get_jobs

if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger('pychemia')
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.DEBUG)

    description = """Launch VASP for non-evaluated entries in a PyChemia Database"""

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-t', '--host',
                        default='localhost', metavar='server', type=str,
                        help='Hostname or address (default: localhost)')
    parser.add_argument('-o', '--port',
                        default=27017, metavar='port', type=int,
                        help='MongoDB port (default: 27017)')
    parser.add_argument('-u', '--user',
                        default=None, metavar='username', type=str,
                        help='Username (default: None)')
    parser.add_argument('-p', '--passwd',
                        default=None, metavar='password', type=str,
                        help='Password (default: None)')
    parser.add_argument('-d', '--dbname',
                        default=None, metavar='dbname', type=str, nargs='+',
                        help='PyChemia Database name (default: None)')
    parser.add_argument('-b', '--binary',
                        default='vasp', metavar='path', type=str,
                        help='VASP binary (default: vasp)')
    parser.add_argument('-s', '--source_dir',
                        default=None, metavar='path', type=str, nargs='+',
                        help='Source Directory where KPOINTS, POSCAR, INCAR and POTCAR are (default: None)')
    parser.add_argument('-r', '--replicaset',
                        default=None, metavar='name', type=str,
                        help='ReplicaSet  (default: None)')
    parser.add_argument('--ssl', action='store_true',
                        help='Use SSL to connect to MongoDB  (default: No)')
    parser.add_argument('--pbs_ppn',
                        default=1, metavar='N', type=int,
                        help='Number of MPI parallel processes (default: 1)')
    parser.add_argument('--pbs_mail',
                        default=None, metavar='user@mail.server', type=str,
                        help='Mail address for PBS (default: None)')
    parser.add_argument('--pbs_queue',
                        default=None, metavar='queue_name', type=str,
                        help='Queue for PBS (default: None)')
    parser.add_argument('--pbs_nhours',
                        default=24, metavar='N', type=int,
                        help='Number of hours for PBS (default: 24)')
    parser.add_argument('--pbs_user',
                        default=None, metavar='username', type=str,
                        help='Username for PBS (default: None)')

    args = parser.parse_args()
    if args.dbname is None:
        parser.print_help()
        exit(1)

    print(args)
    db_settings = {'host': args.host, 'port': args.port, 'ssl': args.ssl, 'replicaset': args.replicaset}
    if args.user is not None:
        if args.passwd is None:
            raise ValueError('Password is mandatory if user is entered')
        db_settings['user'] = args.user
        db_settings['passwd'] = args.passwd
    print('pyChemia Evaluator using VASP')
    print('')
    print('dbname    : %s' % args.dbname)
    print('source_dir: %s' % args.source_dir)
    print('')
    print('host      : %s' % args.host)
    print('port      : %d' % args.port)
    print('user      : %s' % args.user)
    print('replicaset: %s' % args.replicaset)
    print('binary    : %s' % str(args.binary))
    print('ssl       : %s' % str(args.ssl))

    assert(len(args.dbname) == len(args.source_dir))

    while True:

        for idb in range(len(args.dbname)):

            print('DATABASE: %s' % args.dbname[idb])
            db_settings['name'] = args.dbname[idb]
            pcdb = pychemia.db.get_database(db_settings)
            popu = pychemia.population.NonCollinearMagMoms(pcdb, source_dir=args.source_dir[idb])

            print('Number of candidates evaluated: %d' % len(popu.actives_evaluated))
            print('Number of candidates not evaluated: %d' % len(popu.actives_no_evaluated))

            to_compute = popu.actives_no_evaluated

            print('Candidates to compute:')
            for ijob in to_compute:
                print(ijob)
            jobs = get_jobs(args.pbs_user)
            jobnames = [jobs[x]['Job_Name'] for x in jobs]
            source_dir = args.source_dir[idb]

            for ijob in to_compute:
                if str(ijob) not in jobnames:
                    data_collected = popu.collect_data(ijob, workdir=source_dir + os.sep + str(ijob))
                    if not data_collected:
                        print('Preparing and submitting job: %s' % str(ijob))
                        popu.prepare_folder(ijob, workdir=source_dir + os.sep + str(ijob))

                        pbs = pychemia.runner.PBSRunner(source_dir + os.sep + str(ijob))
                        pbs.initialize(ppn=args.pbs_ppn, walltime=[args.pbs_nhours, 0, 0], mail=args.pbs_mail,
                                       queue=args.pbs_queue)
                        pbs.set_template(source_dir+os.sep+'template.pbs')
                        pbs.write_pbs()
                        pbs.submit()
                else:
                    print('Job %s is on queue or running' % str(ijob))
        print('I will be waiting for 60 minutes before checking new candidates')
        time.sleep(3600)



from collections.abc import Mapping
from pychemia.utils.periodic import *

madelung_exceptions = {'Cr': ['Ar', '4s1', '3d5'],
                       'Cu': ['Ar', '4s1', '3d10'],
                       'Nb': ['Kr', '5s1', '4d4'],
                       'Mo': ['Kr', '5s1', '4d5'],
                       'Ru': ['Kr', '5s1', '4d7'],
                       'Rh': ['Kr', '5s1', '4d8'],
                       'Pd': ['Kr', '4d10'],
                       'Ag': ['Kr', '5s1', '4d10'],
                       'La': ['Xe', '6s2', '5d1'],
                       'Ce': ['Xe', '6s2', '4f1', '5d1'],
                       'Gd': ['Xe', '6s2', '4f7', '5d1'],
                       'Pt': ['Xe', '6s1', '4f14', '5d9'],
                       'Au': ['Xe', '6s1', '4f14', '5d10'],
                       'Ac': ['Rn', '7s2', '6d1'],
                       'Th': ['Rn', '7s2', '6d2'],
                       'Pa': ['Rn', '7s2', '5f2', '6d1'],
                       'U': ['Rn', '7s2', '5f3', '6d1'],
                       'Np': ['Rn', '7s2', '5f4', '6d1'],
                       'Cm': ['Rn', '7s2', '5f7', '6d1'],
                       'Lr': ['Rn', '7s2', '5f14', '7p1']}


class Element:

    def __init__(self, value=None):

        if value in atomic_symbols:
            self.symbol = value
        elif value.capitalize() in atomic_symbols:
            self.symbol = value.capitalize()
        else:
            raise ValueError('Symbol %s does not appear on the periodic table' % value)

    @property
    def name(self):
        return atomic_name(self.symbol)

    @property
    def atomic_number(self):
        return atomic_number(self.symbol)

    @property
    def group(self):
        return group(self.symbol)

    @property
    def period(self):
        return period(self.symbol)

    @property
    def block(self):
        return block(self.symbol)

    @property
    def valence(self):
        return valence(self.symbol)

    @property
    def valence_nominal(self):
        return valence_nominal(self.symbol)

    @property
    def mass(self):
        return mass(self.symbol)

    @property
    def covalent_radius(self):
        return covalent_radius(self.symbol)

    @property
    def electronegativity(self):
        return electronegativity(self.symbol)

    @property
    def crystal_structure(self):
        return crystal_structure(self.symbol)

    @property
    def phase(self):
        return phase(self.symbol)

    @property
    def boiling_point(self):
        return boiling_point(self.symbol)

    @property
    def melting_point(self):
        return melting_point(self.symbol)

    @property
    def oxidation_states(self):
        return oxidation_state(self.symbol)

    @property
    def oxidation_states_common(self):
        return oxidation_state(self.symbol, common=True)

    def __str__(self):

        ret = """
Symbol: %s
Name  : %s       

Z     : %d
Group : %d
Period: %d
Block : %s

Valence          : %f
Valence (Nominal): %f

Mass              : %f
Covalent Radius   : %f
Electronegativity : %f

Crystal Structure : %s
Phase             : %s
Boiling Point     : %f
Melting Point     : %f
        """ % (self.symbol, self.name, self.atomic_number, self.group, self.period, self.block,
               self.valence, self.valence_nominal,
               self.mass, self.covalent_radius, self.electronegativity, self.crystal_structure, self.phase,
               self.boiling_point, self.melting_point)
        return ret

    def previous_inert(self):

        inerts = ['He', 'Ne', 'Ar', 'Kr', 'Xe', 'Rn', 'Og']

        if self.period == 1:
            return None
        else:
            # In case the element is already a noble gas the previous one look one period above in Periodic Table
            return inerts[self.period - 2]

    @property
    def madelung_filling(self):

        order = ['1s', '2s', '2p', '3s', '3p', '4s', '3d', '4p', '5s', '4d',
                 '5p', '6s', '4f', '5d', '6p', '7s', '5f', '6d', '7p', '8s',
                 '5g', '6f', '7d', '8p', '9s']

        # We start with the total number of electron and get those associated to orbitals following the order
        capacities = {}
        # l quantum number
        for lqn in range(4):
            label = Element.orbital_label_from_number(lqn)
            nele = Element.max_electrons_subshell(label)
            capacities[label] = nele

        max_electrons = {}
        for ishell in order:
            label = ishell[-1]
            maxele = Element.max_electrons_subshell(label)
            max_electrons[ishell] = maxele

        ret = []
        if self.previous_inert() is not None:
            inert = self.__class__(self.previous_inert())
            ret.append(inert.symbol)
            inert_remain = inert.atomic_number
            # Consume the shells up to the previous inert atom
            numele = self.atomic_number - inert.atomic_number
        else:
            numele = self.atomic_number
            inert_remain = 0

        for i in order:
            if inert_remain >= max_electrons[i]:
                inert_remain -= max_electrons[i]
            elif inert_remain == 0:
                if numele >= max_electrons[i]:
                    numele -= max_electrons[i]
                    ret.append(i + str(max_electrons[i]))
                elif numele == 0:
                    break
                elif numele < max_electrons[i]:
                    ret.append(i + str(numele))
                    break
        return ret

    @staticmethod
    def azimuthal_quantum_number(label):
        aqn = {'s': 0, 'p': 1, 'd': 2, 'f': 3, 'g': 4}
        if label not in ['s', 'p', 'd', 'f', 'g']:
            raise ValueError('Not such label for an orbital: %s' % label)
        return aqn[label]

    # lqn = angular momemtum
    @staticmethod
    def orbital_label_from_number(lqn):
        orbitals = ['s', 'p', 'd', 'f', 'g']
        if lqn not in range(4):
            raise ValueError('Not such azimuthal quantum number: %s' % lqn)
        return orbitals[lqn]

    @staticmethod
    def max_electrons_subshell(subshell):

        if subshell in ['s', 'p', 'd', 'f', 'g']:
            ll = Element.azimuthal_quantum_number(subshell)
        elif subshell in [0, 1, 2, 3, 4]:
            ll = subshell
        else:
            raise ValueError('Not a valid subshell: %s' % subshell)
        return 2 * (2 * ll + 1)

    @property
    def electronic_configuration(self):
        """
        Return the known electronic configuration including exceptions to Madelung's rule
        Based on:

        https://en.wikipedia.org/wiki/Electron_configuration#Atoms:_Aufbau_principle_and_Madelung_rule

        :return:
        """
        if self.symbol in madelung_exceptions:
            return madelung_exceptions[self.symbol]
        else:
            return self.madelung_filling

    @property
    def is_madelung_exception(self):

        return self.symbol in madelung_exceptions



import os
import sys
from pychemia import HAS_PYMATGEN, pcm_log
from .structure import Structure
from pychemia.code.vasp import read_poscar
from pychemia.code.abinit import AbinitInput


def structure_from_file(structure_file):
    """
    Attempts to reconstruct a PyChemia Structure from the contents of any given file. Valid entries

    :param structure_file: The path to a file where the structure can be reconstructed
    :type structure_file: str
    :return: PyChemia Structure if succeed, None otherwise
    """
    st = None
    basename = os.path.basename(structure_file)
    if not os.path.isfile(structure_file):
        raise ValueError("ERROR: Could not open file '%s'" % structure_file)
    if basename[-4:].lower() == 'json':
        st = Structure.load_json(structure_file)
    elif basename[-3:].lower() == 'cif' and HAS_PYMATGEN:
        import pychemia.external.pymatgen
        st = pychemia.external.pymatgen.cif2structure(structure_file)[0]
    elif 'poscar' in basename.lower():
        st = read_poscar(structure_file)
    elif 'contcar' in basename.lower():
        st = read_poscar(structure_file)
    elif 'abinit' in basename.lower():
        av = AbinitInput(structure_file)
        st = av.get_structure()
    else:
        try:
            st = read_poscar(structure_file)
        except ValueError:
            raise ValueError('Ćould not convert file as POSCAR')
    if st is None:
        pcm_log.debug("ERROR: Could not extract structure from file '%s'" % structure_file)
    return st



from .convergence import ConvergenceCutOffEnergy, ConvergenceKPointGrid
from .polarization import Polarization
from .relax import IonRelaxation
from .static import StaticCalculation
from .stregth import IdealStrength
from .elastic import ElasticModuli
from .relax2 import IonRelaxation2

__author__ = 'Guillermo Avendano-Franco'



#!D:\bld\pychemia_1592382025033\_h_env\python.exe
import os
import sys
import pychemia.code.abinit


if __name__ == '__main__':

    # Script starts from here
    if len(sys.argv) < 2:
        print('No file specified.')
        sys.exit()

    filename = sys.argv[1]
    assert (os.path.isfile(filename))
    iv = pychemia.code.abinit.xyz2input(filename)
    av = pychemia.code.abinit.AbinitInput()
    av.variables = iv.variables
    av.write(filename + ".in")



import random
import uuid
from math import gcd
import numpy as np
from ._population import Population
from pychemia import Composition, Structure, pcm_log
from pychemia.analysis import StructureAnalysis, StructureChanger, StructureMatch
from pychemia.analysis.splitting import SplitMatch
from pychemia.utils.mathematics import unit_vector
from pychemia.utils.periodic import atomic_number, covalent_radius
from pymongo import ASCENDING
from pychemia.db import get_database
from pychemia.crystal import CrystalSymmetry


class RelaxStructures(Population):

    def evaluate_entry(self, entry_id):
        pass

    def __init__(self, name, composition=None, tag='global', target_forces=1E-3, value_tol=1E-2,
                 distance_tolerance=0.3, min_comp_mult=2, max_comp_mult=8, pcdb_source=None, pressure=0.0,
                 target_stress=None, target_diag_stress=None, target_nondiag_stress=None):
        """
        Defines a population of PyChemia Structures,

        The 'name' of the database is used to create the MongoDB database and the structures are
        uniform in composition. A specific 'tag' could be attached to differentiate
        the other instances running concurrently. The 'delta' argument is the scaling
        factor for changers and mixers. In the case of populations supported on
        PyChemia databases the 'new' will erase the database

        :param name: The name of the population. ie the name of the database
        :param composition: The composition uniform for all the members
        :param tag: A tag to differentiate different instances running concurrently
        :return: A new StructurePopulation object
        """
        if composition is not None:
            self.composition = Composition(composition)
        else:
            self.composition = None
        self.tag = tag
        self.target_forces = target_forces
        self.value_tol = value_tol
        self.min_comp_mult = min_comp_mult
        self.max_comp_mult = max_comp_mult
        self.pcdb_source = pcdb_source
        self.pressure = pressure
        if target_stress is None:
            self.target_stress = target_forces
        else:
            self.target_stress = target_stress
        if target_diag_stress is None:
            self.target_diag_stress = self.target_stress
        else:
            self.target_diag_stress = target_diag_stress
        if target_diag_stress is None:
            self.target_nondiag_stress = self.target_stress
        else:
            self.target_nondiag_stress = target_nondiag_stress
        self.name = name
        Population.__init__(self, name, tag, distance_tolerance=distance_tolerance)

        if self.pcdb_source is not None:
            self.sources = {}
            for i in range(min_comp_mult, max_comp_mult+1):
                self.sources[i] = []
                for entry in self.pcdb_source.entries.find({'structure.natom': i*self.composition.natom,
                                                            'structure.nspecies': self.composition.nspecies},
                                                           {'_id': 1}):
                    self.sources[i].append(entry['_id'])

    def recover(self):
        data = self.get_population_info()
        if data is not None:
            self.distance_tolerance = data['distance_tol']
            self.value_tol = data['value_tol']
            self.name = data['name']
            self.target_forces = data['target_forces']

    def get_structure(self, entry_id):
        entry = self.get_entry(entry_id)
        return Structure.from_dict(entry['structure'])

    @staticmethod
    def new_identifier():
        return str(uuid.uuid4())[-12:]

    def new_entry(self, structure, active=True):
        properties = {'forces': None, 'stress': None, 'energy': None}
        status = {self.tag: active, 'tag': self.tag}
        entry = {'structure': structure.to_dict, 'properties': properties, 'status': status}
        entry_id = self.insert_entry(entry)
        pcm_log.debug('Added new entry: %s with tag=%s: %s' % (str(entry_id), self.tag, str(active)))
        return entry_id

    def get_max_force_stress(self, entry_id):
        entry = self.get_entry(entry_id, projection={'properties': 1})
        max_force = None
        max_diag_stress = None
        max_nondiag_stress = None
        if entry is not None and entry['properties'] is not None:
            properties = entry['properties']
            if 'forces' in properties and 'stress' in properties:
                if properties['forces'] is not None and properties['stress'] is not None:
                    forces = np.array(entry['properties']['forces'])
                    stress = np.array(entry['properties']['stress'])
                    max_force = np.max(np.apply_along_axis(np.linalg.norm, 1, forces))
                    max_diag_stress = np.max(np.abs(stress[:3]))
                    max_nondiag_stress = np.max(np.abs(stress[4:]))

        return max_force, max_diag_stress, max_nondiag_stress

    def is_evaluated(self, entry_id):
        max_force, max_diag_stress, max_nondiag_stress = self.get_max_force_stress(entry_id)
        if max_force is None or max_diag_stress is None or max_nondiag_stress is None:
            return False
        elif max_force < self.target_forces and max_diag_stress < self.target_diag_stress + self.pressure:
            if max_nondiag_stress < self.target_nondiag_stress:
                return True
            else:
                return False
        else:
            return False

    def add_random(self, random_probability=0.3):
        """
        Add one random structure to the population
        """
        entry_id = None
        structure = Structure()
        if self.composition is None:
            raise ValueError('No composition associated to this population')
        factor = np.random.randint(self.min_comp_mult, self.max_comp_mult + 1)
        comp = self.composition.composition.copy()
        # print("Initial composition: %s" % comp)
        # print(Composition(comp))
        # print(Composition(comp).symbols)
        for i in comp:
            comp[i] *= factor
        new_comp = Composition(comp)

        while True:
            rnd = random.random()
            condition = {'structure.nspecies': new_comp.nspecies,
                         'structure.natom': new_comp.natom}
            if self.pcdb_source is None:
                rnd = 0
            elif len(self.sources[factor]) == 0:
                rnd = 0
            if self.pcdb_source is None or rnd < random_probability:
                pcm_log.debug('Random Structure')
                structure = Structure.random_cell(new_comp, method='stretching', stabilization_number=5, nparal=5,
                                                  periodic=True)
                break
            else:
                pcm_log.debug('From source')
                entry_id = self.sources[factor][np.random.randint(0, len(self.sources[factor]))]
                structure = self.pcdb_source.get_structure(entry_id)
                print("chosen structure from database =", structure)
                sym = CrystalSymmetry(structure)

                scale_factor = float(np.max(covalent_radius(new_comp.species)) /
                                     np.max(covalent_radius(structure.species)))
                reduce_scale = scale_factor ** (1. / 3)    # WIH
                msg = 'Mult: %d natom: %d From source: %s Spacegroup: %d Scaling: %7.3f'
                print(msg % (factor, structure.natom, structure.formula, sym.number(), scale_factor))
                # structure.set_cell(np.dot(scale_factor * np.eye(3), structure.cell)) # WIH
                structure.set_cell(np.dot(reduce_scale * np.eye(3), structure.cell))  # WIH
                print("symbols before change = ", structure.symbols)
                structure.symbols = new_comp.symbols
                print("symbols after change = ", structure.symbols)
                self.sources[factor].remove(entry_id)
                break

        return self.new_entry(structure), entry_id

    def check_duplicates(self, ids):
        """
        Computes duplicate structures measuring its distance when their value is larger than value_tol.
        If the distance is lower than 'distance_tol' the structures will be cosidered as duplicates.

        :param ids:
        :return: (dict) Dictionary of duplicates, the keys are the ids of the duplicates and the value is the structure
                        from which the structure is duplicated. In general the energy of the 'value' is lower than the
                        'key'

        """

        ret = {}
        selection = self.ids_sorted(ids)
        values = np.array([self.value(i) for i in selection])
        if len(values) == 0:
            return ret
        diffs = np.ediff1d(values)

        for i in range(len(diffs)):
            idiff = diffs[i]
            if idiff < self.value_tol:
                ident1 = selection[i]
                ident2 = selection[i + 1]
                pcm_log.debug('Testing distances between %s and %s' % (str(ident1), str(ident2)))
                distance = self.distance(ident1, ident2)
                # print 'Distance = ', distance
                if distance < self.distance_tolerance:
                    pcm_log.debug('Distance %7.3f < %7.3f' % (distance, self.distance_tolerance))
                    ret[ident2] = ident1
        if len(ret) > 0:
            pcm_log.debug('Number of duplicates %d' % len(ret))
        return ret

    def get_duplicates(self, ids, tolerance, fast=False):
        dupes_dict = {}
        dupes_list = []
        values = {}
        for i in ids:
            values[i] = self.value(i)
        selection = self.ids_sorted(ids)
        print('Searching duplicates in %d structures' % len(selection))
        for i in range(len(selection) - 1):
            entry_id = selection[i]
            value_i = values[entry_id]
            for j in range(i + 1, len(selection)):
                entry_jd = selection[j]
                if fast and entry_jd in dupes_list:
                    continue
                value_j = values[entry_jd]
                if abs(value_i - value_j) < self.value_tol:
                    distance = self.distance(entry_id, entry_jd)
                    if distance < tolerance:
                        if entry_id in dupes_dict:
                            dupes_dict[entry_id].append(entry_jd)
                        else:
                            dupes_dict[entry_id] = [entry_jd]
                        dupes_list.append(entry_jd)
        return dupes_dict, [x for x in selection if x in dupes_list]

    def cleaned_from_duplicates(self, ids):
        selection = self.ids_sorted(ids)
        duplicates_dict = self.check_duplicates(selection)
        return [x for x in selection if x not in duplicates_dict.keys()]

    def diff_values_matrix(self):

        members = self.members
        ret = np.zeros((len(members), len(members)))

        for i in range(len(members)):
            for j in range(i, len(members)):

                if self.value(members[i]) is not None and self.value(members[j]) is not None:
                    ret[i, j] = np.abs(self.value(members[i]) - self.value(members[j]))
                else:
                    ret[i, j] = float('nan')
                ret[j, i] = ret[i, j]
        return ret

    def distance(self, entry_id, entry_jd, rcut=50):

        ids_pair = [entry_id, entry_jd]
        ids_pair.sort()
        distance_entry = self.pcdb.db.distances.find_one({'pair': ids_pair}, {'distance': 1})
        self.pcdb.db.distances.create_index([("pair", ASCENDING)])

        if distance_entry is None:
            print('Distance not in DB')
            fingerprints = {}
            for entry_ijd in [entry_id, entry_jd]:

                if self.pcdb.db.fingerprints.find_one({'_id': entry_ijd}) is None:
                    structure = self.get_structure(entry_ijd)
                    analysis = StructureAnalysis(structure, radius=rcut)
                    x, ys = analysis.fp_oganov()
                    fingerprint = {'_id': entry_ijd}
                    for k in ys:
                        atomic_number1 = atomic_number(structure.species[k[0]])
                        atomic_number2 = atomic_number(structure.species[k[1]])
                        pair = '%06d' % min(atomic_number1 * 1000 + atomic_number2,
                                            atomic_number2 * 1000 + atomic_number1)
                        fingerprint[pair] = list(ys[k])

                    if self.pcdb.db.fingerprints.find_one({'_id': entry_ijd}) is None:
                        self.pcdb.db.fingerprints.insert(fingerprint)
                    else:
                        self.pcdb.db.fingerprints.update({'_id': entry_ijd}, fingerprint)
                    fingerprints[entry_ijd] = fingerprint
                else:
                    fingerprints[entry_ijd] = self.pcdb.db.fingerprints.find_one({'_id': entry_ijd})

            dij = []
            for pair in fingerprints[entry_id]:
                if pair in fingerprints[entry_jd] and pair != '_id':
                    uvect1 = unit_vector(fingerprints[entry_id][pair])
                    uvect2 = unit_vector(fingerprints[entry_jd][pair])
                    dij.append(0.5 * (1.0 - np.dot(uvect1, uvect2)))
            distance = float(np.mean(dij))
            self.pcdb.db.distances.insert({'pair': ids_pair, 'distance': distance})
        else:
            distance = distance_entry['distance']
        return distance

    def add_from_db(self, db_settings, sizemax=1):
        if self.composition is None:
            raise ValueError('No composition associated to this population')
        comp = Composition(self.composition)
        readdb = get_database(db_settings)

        index = 0
        for entry in readdb.entries.find({'structure.formula': comp.formula,
                                          'structure.natom': {'$lte': self.min_comp_mult * comp.natom,
                                                              '$gte': self.max_comp_mult * comp.natom}}):
            if index < sizemax:
                print('Adding entry ' + str(entry['_id']) + ' from ' + readdb.name)
                self.new_entry(readdb.get_structure(entry['_id']))
                index += 1

    def move_random(self, entry_id, factor=0.2, in_place=False, kind='move'):
        structure = self.get_structure(entry_id)
        changer = StructureChanger(structure=structure)
        if kind == 'move':
            changer.random_move_many_atoms(epsilon=factor)
        else:  # change
            changer.random_change(factor)
        if in_place:
            return self.set_structure(entry_id, changer.new_structure)
        else:
            return self.new_entry(changer.new_structure, active=False)

    def move(self, entry_id, entry_jd, factor=0.2, in_place=False):
        """
        Moves entry_id in the direction of entry_jd
        If in_place is True the movement occurs on the
        same address as entry_id

        :param factor:
        :param entry_id:
        :param entry_jd:
        :param in_place:
        :return:
        """
        structure_mobile = self.get_structure(entry_id)
        structure_target = self.get_structure(entry_jd)

        if structure_mobile.natom != structure_target.natom:
            # Moving structures with different number of atoms is only implemented for smaller structures moving
            # towards bigger ones by making a super-cell and only if their size is smaller that 'max_comp_mult'

            mult1 = structure_mobile.get_composition().gcd
            mult2 = structure_target.get_composition().gcd
            lcd = mult1 * mult2 / gcd(mult1, mult2)
            if lcd > self.max_comp_mult:
                # The resulting structure is bigger than the limit
                # cannot move
                if not in_place:
                    return self.new_entry(structure_mobile)
                else:
                    return entry_id

        # We will move structure1 in the direction of structure2
        match = StructureMatch(structure_target, structure_mobile)
        match.match_size()
        match.match_shape()
        match.match_atoms()
        displacements = match.reduced_displacement()

        new_reduced = match.structure2.reduced + factor * displacements
        new_cell = match.structure2.cell
        new_symbols = match.structure2.symbols
        new_structure = Structure(reduced=new_reduced, symbols=new_symbols, cell=new_cell)
        if in_place:
            return self.set_structure(entry_id, new_structure)
        else:
            return self.new_entry(new_structure, active=False)

    def __str__(self):
        ret = '\n'
        ret += '[%s] Population type:    %s\n' % (self.tag, 'Relax Structures')
        ret += '[%s] Database:           %s\n' % (self.tag, self.name)
        ret += '[%s] Tag:                %s\n' % (self.tag, self.tag)
        ret += '[%s] Target-Forces:      %7.2E\n' % (self.tag, self.target_forces)
        ret += '[%s] Value tolerance:    %7.2E\n' % (self.tag, self.value_tol)
        ret += '[%s] Distance tolerance: %7.2E\n\n' % (self.tag, self.distance_tolerance)
        if self.composition is not None:
            ret += '[%s] Composition:                    %s\n' % (self.tag, self.composition.formula)
            ret += '[%s] Minimal composition multiplier: %d\n' % (self.tag, self.min_comp_mult)
            ret += '[%s] Maximal composition multiplier: %d\n' % (self.tag, self.max_comp_mult)
        ret += '[%s] Members:            %d\n' % (self.tag, len(self.members))
        ret += '[%s] Actives:            %d\n' % (self.tag, len(self.actives))
        ret += '[%s] Evaluated:          %d\n' % (self.tag, len(self.evaluated))
        ret += '\n'
        return ret

    def value(self, entry_id):
        entry = self.get_entry(entry_id)
        structure = self.get_structure(entry_id)
        if 'properties' not in entry:
            pcm_log.debug('This entry has no properties %s' % str(entry['_id']))
            return None
        elif entry['properties'] is None:
            return None
        elif 'energy' not in entry['properties']:
            pcm_log.debug('This entry has no energy in properties %s' % str(entry['_id']))
            return None
        else:
            return entry['properties']['energy'] / structure.get_composition().gcd

    @property
    def to_dict(self):
        return {'name': self.name,
                'tag': self.tag,
                'target_forces': self.target_forces,
                'value_tol': self.value_tol,
                'distance_tolerance': self.distance_tolerance}

    def from_dict(self, population_dict):
        return RelaxStructures(name=population_dict['name'],
                               tag=population_dict['tag'],
                               target_forces=population_dict['target_forces'],
                               value_tol=population_dict['value_tol'],
                               distance_tolerance=population_dict['distance_tolerance'])

    def cross(self, ids):

        assert len(ids) == 2

        structure1 = self.get_structure(ids[0])
        structure2 = self.get_structure(ids[1])

        split_match = SplitMatch(structure1, structure2)
        st1, st2 = split_match.get_simple_match()

        entry_id = self.new_entry(st1, active=True)
        entry_jd = self.new_entry(st2, active=True)

        return entry_id, entry_jd

    def str_entry(self, entry_id):

        struct = self.get_structure(entry_id)
        return str(struct)




import os
import socket
import time
import pychemia
from multiprocessing import Process
from pychemia import pcm_log, HAS_PYMONGO

if HAS_PYMONGO:
    from pychemia.db import get_database


class FireballCollector:

    def __init__(self, dbname, source_dir=None, source_file=None, output_names=None, output_file=None,
                 db_settings=None, nconcurrent=1):
        """
        FireballCollector is a class collect Fireball executions into one PyChemia Database.

        The name of the database will be 'dbname' and executions will be processed from a give 'source_dir', a path
        to search for Fireball calculations or 'source_file', one file with one directory per line where Fireball
        executions are stored.
        The variable 'output_names' sets possible output files for Fireball. Typically Fireball output is the standard
        output and the user has to redirect the output to a file with an arbitrary name.

        The processing of directories happens in parallel using 'nconcurrent' python processes.
        By default the database is created on the local machine assuming no authorization or authentication, otherwise
        use the dictionary 'db_settings' to get more control on the database location and access.

        :param dbname:      (str) Name of the database that will contain the Fireball executions
        :param source_dir:  (str) Path for a directory that will be explored recursevely for suitable Fireball
                                  executions
        :param source_file: (str) Path to a file describing directories with Fireball executions, one per line
        :param output_names: (list) List of possible output filenames
        :param output_file: (str) Path to a file with possible output names, one per line
        :param db_settings: (dict) Extra parameters for locating and accessing the mongo server.
        :param nconcurrent: Number of concurrent python executions to fill the database (default: 1)
        """
        self.db_settings = db_settings
        self.dbname = dbname
        self.source_dir = source_dir
        self.source_file = source_file
        self.output_names = output_names
        self.output_file = output_file
        self.nconcurrent = nconcurrent
        self.sleeping_time = 2
        self.database = None

        if self.output_file is not None and os.path.exists(self.output_file):
            rf = open(self.output_file)
            self.output_names = [x.strip() for x in rf.readlines()]

        if self.source_dir is None and self.source_file is None:
            raise ValueError("One of the variables source_dir or source_file need to be defined")

        if self.db_settings is None:
            self.db_settings = {'name': self.dbname}
        else:
            self.db_settings['name'] = self.dbname

        self.db_settings = dict(self.db_settings)
        # The first component of each pair in to_evaluate is the name of the database
        self.db_settings['name'] = self.dbname

        self.database = get_database(self.db_settings)

    def worker(self, path):
        """
        From a given 'path', the worker will collect information from several files
        such as:

        fireball.in
        eigen.dat
        param.dat
        answer.bas
        output collected from the standard output of fireball.x


        After collecting data into dictionaries, the data is stored in a PyChemia database
        """

        pcdb = get_database(self.db_settings)
        files = os.listdir(path)
        properties = {}
        geo = None

        if os.path.lexists(path+os.sep+'fireball.in'):
            properties['path'] = path
            score = 'input'
            try:
                invars = pychemia.code.fireball.read_fireball_in(path + os.sep + 'fireball.in')
                properties['input'] = invars
            except:
                print('Bad fireball.in on %s' % path)

            if os.path.exists(path+os.sep+'param.dat'):
                try:
                    score += '+param'
                    param = pychemia.code.fireball.read_param(path + os.sep + 'param.dat')
                    properties['param_dat'] = param
                except:
                    print('Bad param.dat')

            if os.path.lexists(path+os.sep+'eigen.dat'):
                try:
                    score += '+eigen'
                    eigen = pychemia.code.fireball.read_eigen(path + os.sep + 'eigen.dat')
                    properties['eigen_dat'] = eigen
                except:
                    print('bad eigen.dat')

            if os.path.lexists(path+os.sep+'answer.bas'):
                try:
                    score += '+answer'
                    geo = pychemia.code.fireball.read_geometry_bas(path + os.sep + 'answer.bas')
                    periodic = False
                except:
                    print('Bad answer.bas')

            for ioutput in self.output_names:
                if os.path.isfile(path + os.sep + ioutput):
                    rf = open(path + os.sep + ioutput)
                    line = rf.readline()
                    rf.close()
                    if "Welcome to FIREBALL" in line:
                        try:
                            output = pychemia.code.fireball.read_final_fireball_relax(path + os.sep + ioutput)
                            properties['output'] = output
                            break
                        except:
                            print('Bad output %s on %s' % (ioutput, path))

            for ifile in files:
                if ifile[-3:] == 'lvs':
                    try:
                        cell = pychemia.code.fireball.read_lvs(path + os.sep + ifile)
                        periodic = True
                        lat = pychemia.crystral.Lattice(cell)
                        if lat.volume > 1E7:
                            print('Lattice too big, assuming non-periodic structure')
                            periodic = False
                    except:
                        print('Bad %s' % ifile)

        if geo is not None:
            print('DB: %d entries, Path: %s' % (pcdb.entries.count(), path))
            if periodic:
                st = pychemia.Structure(symbols=geo.symbols, positions=geo.positions, cell=cell)
            else:
                st = pychemia.Structure(symbols=geo.symbols, positions=geo.positions, periodicity=False)

            pcdb.insert(st, properties)
        return properties

    def process_directory(self, path, fireball_dirs):
        """
        Search for directories with fireball.in and answer.bas and add them to the
        current list of 'fireball_dirs'

        The search is recursive to subdirectories.
        """
        files = os.listdir(path)
        score = False
        if os.path.lexists(path+os.sep+'fireball.in') and os.path.exists(path+os.sep+'answer.bas'):
            for ioutput in self.output_names:
                if os.path.isfile(path + os.sep + ioutput):
                    rf = open(path + os.sep + ioutput)
                    if "Welcome to FIREBALL" in rf.readline():
                        score = True
                    rf.close()

        if score:
            print(path)
            ret.append(path)

        for ifile in files:
            if os.path.isdir(path + os.sep + ifile):
                self.process_directory(path + os.sep + ifile, fireball_dirs)

    def save_list_candidates(self, filename):
        """
        Scan all databases looking for candidates for evaluation

        :return: A list of pairs, each pair contains the name of the database
                    and candidate identifier.
        """

        ret = []
        if not os.path.isfile(filename):
            self.process_directory(self.source_dir, ret)
            wf = open('fireball_directories.txt', 'w')
            for i in ret:
                wf.write("%s\n" % i)
            wf.close()
        else:
            rf = open(filename)
            ret = [x[:-1] for x in rf.readlines()]
            rf.close()

        return ret

    def is_evaluated(self, pcdb, path):
        if pcdb.entries.find({'properties.path': path}).count() > 0:
            return True
        else:
            return False

    def run(self):
        """
        Process the list of directories from 'source_dir' or 'source_file'
        Using 'nconcurrent' python processes.

        :return:
        """

        procs = []
        ids_running = []

        # Creating a list to store the 'nconcurrent' running jobs
        for i in range(self.nconcurrent):
            procs.append(None)
            ids_running.append(None)

        to_evaluate = []
        if self.source_file is None:
            self.process_directory(self.source_dir, to_evaluate)
        else:
            rf = open(self.source_file)
            to_evaluate = [x.strip() for x in rf.readlines()]

        # Main loop looking permanently for candidates for evaluation
        while True:

            index = 0
            currently_evaluating = 0
            for j in range(self.nconcurrent):
                if procs[j] is not None and procs[j].is_alive():
                    currently_evaluating += 1
            print('Candidates to evaluate: %d  Candidates in evaluation: %d' % (len(to_evaluate), currently_evaluating))

            while index < len(to_evaluate):

                pcdb = self.database

                slot = None
                while True:
                    for j in range(self.nconcurrent):
                        if procs[j] is None or not procs[j].is_alive():
                            slot = j
                            break
                    if slot is None:
                        time.sleep(self.sleeping_time)
                    else:
                        break

                # The function is_evaluated needs two arguments, the database object and entry identifier and
                # must return a boolean to decide if the candidate should be evaluated.
                if not self.is_evaluated(pcdb, to_evaluate[index]):
                    ids_running[slot] = to_evaluate[index]
                    # This is the actual call to the worker, it must be a function with 4 arguments:
                    # The database settings, the entry identifier, the working directory and arguments for the worker
                    print('[Slot: %d] Evaluating: %s' % (slot, to_evaluate[index]))
                    procs[slot] = Process(target=self.worker, args=(to_evaluate[index],))
                    procs[slot].start()
                    time.sleep(0.5)
                else:
                    print('[Slot: %d] Evaluated: %s' % (slot, to_evaluate[index]))
                    pass

                index += 1
            time.sleep(self.sleeping_time)



import os as _os
import shutil as _shutil
import subprocess as _subprocess
from abc import ABCMeta, abstractmethod
from .function import FunctionEvaluator, FunctionObjectiveFunction
from .cluster import cluster_worker, cluster_evaluator, cluster_launcher
from .cluster_fireball import cluster_fb_evaluator, cluster_fb_launcher, cluster_fb_worker
from .direct_evaluator import DirectEvaluator
from .Fireball2PyChemiaDB import FireballCollector


def execute(basedir, command, script):
    """
    Utility that copy a given script and execute the given
    command inside the directory

    :param basedir: (str) Basedir of the directory with script
    :param command: (str) Command to execute
    :param script: (str) Script to call inside basedir
    """
    cwd = _os.getcwd()
    if not _os.path.isfile(basedir + '/' + _os.path.basename(script)):
        _shutil.copy(script, basedir)
    _os.chdir(basedir)
    print('Executing... ' + command + ' ' + script)
    _subprocess.call([command, script])
    _os.chdir(cwd)


class Evaluator:
    __metaclass__ = ABCMeta

    def __init__(self):
        self.population = None

    @abstractmethod
    def initialize(self, population):
        self.population = population

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @property
    @abstractmethod
    def is_running(self):
        pass



import numpy as np


class Verlet:
    def __init__(self, x0, v0, delta, accel_function, dissipation=1.0, niter=None, min_vel=None):

        self.x0 = np.array(x0).reshape((-1, 3))
        self.v0 = np.array(v0).reshape((-1, 3))
        self.delta = delta
        self.accel_function = accel_function
        self.trajectory = [self.x0]
        self.niter = niter
        # Minimal velocity to stop
        self.min_vel = min_vel

    def run(self):

        x0 = self.x0
        x1 = self.x0 + self.v0 * self.delta + 0.5 * self.accel_function(self.x0) * self.delta ** 2
        self.trajectory.append(x1)
        iteration = 0
        while True:
            xn = 2 * x1 - x0 + self.accel_function(x1) * self.delta ** 2
            self.trajectory.append(xn)
            x0 = x1
            x1 = xn
            iteration += 1

            if self.niter is not None:
                if iteration == self.niter:
                    break

            if self.min_vel is not None:
                vel = np.array((x1 - x0) / self.delta)
                if np.max(np.apply_along_axis(np.linalg.norm, 1, vel)) > self.min_vel:
                    break



import os
import re
from .output import VaspOutput


def read_vasp_stdout(filename):
    if not os.path.isfile(filename):
        raise ValueError("Could not read: %s" % filename)
    rf = open(filename)
    data = rf.read()
    rf.close()

    flaws = re.findall('[0-9]*\.[0-9]*E[+-][0-9]*-[0-9]*\.[0-9]*E[+-][0-9]*', data)
    flaws_splitted = re.findall('([0-9]*\.[0-9]*E[+-][0-9]*)(-[0-9]*\.[0-9]*E[+-][0-9]*)', data)
    for i in range(len(flaws)):
        data = data.replace(flaws[i], flaws_splitted[i][0] + ' ' + flaws_splitted[i][1])

    re_str = r'\n([\w]{3})\:\s*([\d]+)+\s*([\dE+-.]+)\s*([\dE+-.]+)\s*([\dE+-.]+)\s*([\d]+)\s*([\dE+-.]+)\s*[-\ .\w+]+'

    result = re.findall(re_str, data)

    ret = []
    for i in result:
        ret.append([i[0], int(i[1]), float(i[2]), float(i[3]), float(i[4]), int(i[5]), float(i[6])])

    number_of_scf_per_ionic_iter = []
    final_energy_after_scf = []
    index = 0
    counter = 0
    for i in ret:
        if i[1] > index:
            index = i[1]
        else:
            number_of_scf_per_ionic_iter.append(index)
            final_energy_after_scf.append(ret[counter][2])
            index = 0
        counter += 1

    return {'iterations': number_of_scf_per_ionic_iter, 'energies': final_energy_after_scf, 'data': ret}



import os
import time
import numpy as np
from threading import Thread
from ase.calculators.lj import LennardJones
from ase.optimize import QuasiNewton
from ase.constraints import FixAtoms, UnitCellFilter
import pychemia
import pychemia.external.ase
from pychemia.utils.serializer import generic_serializer


class AseObjectiveFunction:
    def __init__(self):
        self.population = None

    def initialize(self, population):
        self.population = population

    def ids_sorted(self, selection):
        values = np.array([self.population.value(i) for i in selection])
        argsort = np.argsort(values)
        return np.array(selection)[argsort]

    def get_values(self, selection):
        ret = {}
        for i in selection:
            ret[i] = self.population.value(i)
        return ret


class AseEvaluator:
    def __init__(self):
        self.process = None
        self.thread = None
        self.population = None

    def initialize(self, population):
        self.population = population

    def evaluate(self, imember):
        entry = self.population.get_entry(imember)
        pcm_structure = pychemia.Structure.from_dict(entry['structure'])

        ase_structure = pychemia.external.ase.pychemia2ase(pcm_structure)
        ase_structure.set_calculator(LennardJones())

        dyn = QuasiNewton(ase_structure)
        dyn.run()

        ase_structure.set_constraint(FixAtoms(mask=[True for atom in ase_structure]))
        ucf = UnitCellFilter(ase_structure)
        qn = QuasiNewton(ucf)
        qn.run()

        new_structure = pychemia.external.ase.ase2pychemia(ase_structure)
        energy = ase_structure.get_potential_energy()
        forces = ase_structure.get_forces()
        stress = ase_structure.get_stress()
        new_properties = {'energy': float(energy), 'forces': generic_serializer(forces),
                          'stress': generic_serializer(stress)}

        self.population.db.update(imember, structure=new_structure, properties=new_properties)

    @property
    def is_running(self):
        if self.thread is not None:
            return self.thread.is_alive()
        else:
            return False

    def run(self):

        def worker(evaluator, population):
            while True:
                for i in population.actives:
                    if not population.is_evaluated(i):
                        evaluator.evaluate(i)
                time.sleep(1)
                if os.path.exists('stop'):
                    os.remove('stop')
                    return

        # self.process = Process(target=worker, args=(self.population,))
        # self.process.start()
        self.thread = Thread(target=worker, args=(self, self.population,))
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        if self.process is not None and not self.process.is_alive():
            self.process.terminate()



import math
import numpy as np
import scipy.spatial
import itertools
from pychemia.utils.mathematics import integral_gaussian
from pychemia.utils.periodic import atomic_number

__author__ = "Guillermo Avendano-Franco"


class ClusterAnalysis:
    def __init__(self, structure):
        """
        Cluster Analysis provides routines to compute Structure Analysis
        for finite systems such as molecules and clusters.

        :param structure: (pychemia.Structure) A PyChemia Structure object

        """

        self.structure = structure.copy()
        self.structure.relocate_to_cm()
        # This is the maximal distance between any two atoms
        self.max_distance = np.max(self.distance_matrix())

    def distance_matrix(self):
        """
        Compute the distance matrix, the distance between any two particles
        in the entire structure.
        For distance matrices related separated by species, use instead
        distance_matrix_by_species.

        :return:
        """
        return scipy.spatial.distance_matrix(self.structure.positions, self.structure.positions)

    def distance_matrix_by_species(self, specie1, specie2):

        assert self.structure.is_perfect
        group1 = np.array(self.structure.symbols) == specie1
        group2 = np.array(self.structure.symbols) == specie2
        return scipy.spatial.distance_matrix(self.structure.positions[group1], self.structure.positions[group2])

    def all_distances_by_species(self):

        ret = {}
        for i, j in itertools.combinations_with_replacement(self.structure.species, 2):
            pair = tuple(np.sort([i, j]))
            dm = self.distance_matrix_by_species(i, j)
            distances = np.triu(dm, 1).flatten()
            distances = np.sort(distances[distances > 0.0])
            ret[pair] = distances
        return ret

    def discrete_radial_distribution_function(self, delta=0.01, sigma=0.01, integrated=True):
        dist_spec = self.all_distances_by_species()
        discrete_rdf = {}
        nbins = int((self.max_distance + 8 * sigma) / delta)
        if nbins > 10000:
            nbins = 10000
        discrete_rdf_x = np.arange(0, nbins * delta, delta)
        for spec_pair in dist_spec:
            discrete_rdf[spec_pair] = np.zeros(nbins)
            for Rij in dist_spec[spec_pair]:
                # Integrating for bin from - 8*sigma to +8*sigma centered on Rij
                # Values outside this range are negligible
                imin = int(max(0, (Rij - 8 * sigma) / delta))
                imax = int(min(len(discrete_rdf_x), (Rij + 8 * sigma) / delta))
                # log.debug('Computing for distance %7.3f for indices between %d and %d' % (Rij, imin, imax))
                for i in range(imin, imax):
                    x = discrete_rdf_x[i]
                    if not integrated:
                        discrete_rdf[spec_pair][i] += np.exp(-((x - Rij) ** 2) / (2 * sigma * sigma)) / (
                            4 * math.pi * Rij * Rij)
                    else:
                        discrete_rdf[spec_pair][i] += integral_gaussian(x, x + delta, Rij, sigma) / (
                            4 * math.pi * Rij * Rij)

        return discrete_rdf_x, discrete_rdf


class ClusterMatch:
    def __init__(self, structure1, structure2):
        """
        Given two structures, ClusterMatch can compute a one-to-one
        association between atoms of both structures, trying to minimize
        the displacements needed to move one structure into another.

        :param structure1:
        :param structure2:
        """

        assert not structure1.is_periodic
        assert not structure2.is_periodic

        self.structure1 = structure1.copy()
        self.structure2 = structure2.copy()

        self.structure1.canonical_form()
        self.structure2.canonical_form()

        assert self.structure1.symbols == self.structure2.symbols

    def match(self):

        species = self.structure1.species
        species = list(np.array(species)[np.array(atomic_number(species)).argsort()])

        permutation = np.zeros(len(self.structure1.positions), dtype=int)
        num = 0
        for ispecie in species:

            pos1 = self.structure1.positions[np.array(self.structure1.symbols) == ispecie]
            pos2 = self.structure2.positions[np.array(self.structure2.symbols) == ispecie]
            dm = scipy.spatial.distance_matrix(pos1, pos2)

            match_list = np.zeros(len(pos1))
            maxdis = np.max(dm.flatten())
            for i in range(len(pos1)):
                match = dm[i, :].argmin()
                # print " %3s %3d %9.3f %9.3f %9.3f" % (ispecie, match, np.linalg.norm(pos1[i]),
                # np.linalg.norm(pos1[match]), dm[i,match])
                match_list[i] = match
                dm[:, match] = maxdis + 1

            match_list += num
            permutation[num:num + len(pos1)] = match_list
            num += len(pos1)

        # print permutation
        self.structure2.sort_sites_using_list(permutation)



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

# This script download several sets of pseudopotentials
# form the Website of ABINIT
# The pseudopotentials will be located inside the directory
# '.abinit' of the user home directory, just aside of the
# tarballs.
# AE are the Hirshfeld All Electron (AE) densities
# This script is for Python 3 only as it depends on urllib.request
# Not backwards support for Python 2.x

import os
import ftplib
from concurrent.futures import ThreadPoolExecutor
from pychemia.code.abinit import psp_name
from urllib.request import urlopen
import tarfile
import time


def worker(filename, directory, filepath):
    ftp = ftplib.FTP('ftp.abinit.org')  # connect to host, default port
    ftp.login()  # user anonymous, passwd anonymous@
    print('Getting ftp://ftp.abinit.org/%s/%s' % (filepath, filename))
    nofile = True
    while nofile:
        try:
            ftp.retrbinary('RETR ' + filepath + filename, open(directory + '/' + filename, 'wb').write)
            nofile = False
            if os.path.getsize(directory + '/' + filename) == 0:
                os.remove(directory + '/' + filename)
        except ftplib.error_perm:
            print('Retrying download of %s' % filepath)
            ftp.close()
            time.sleep(1)
            if os.path.isfile(directory + '/' + filename):
                os.remove(directory + '/' + filename)
            ftp = ftplib.FTP('ftp.abinit.org')  # connect to host, default port
            ftp.login()  # user anonymous, passwd anonymous@
            nofile = False
        except EOFError:
            print('Retrying download of %s' % filepath)
            ftp.close()
            time.sleep(1)
            if os.path.isfile(directory + '/' + filename):
                os.remove(directory + '/' + filename)
            ftp = ftplib.FTP('ftp.abinit.org')  # connect to host, default port
            ftp.login()  # user anonymous, passwd anonymous@
            nofile = False
    ftp.close()


def get_rpath_psp(kind, exchange, atomicnumber=None):
    """
    Get the ftp path to get the PSP

    :param kind: (str) Source of Pseudopotentials
    :param exchange: (str) 'LDA' or 'GGA'
    :param atomicnumber: (int) Atomic Number
    :return:
    """
    rpath = None
    if kind == 'FHI' and exchange == 'LDA':
        rpath = '/pub/abinitio/Psps/LDA_FHI/'
    elif kind == 'FHI' and exchange == 'GGA':
        rpath = '/pub/abinitio/Psps/GGA_FHI/'
    elif kind == 'GTH' and exchange == 'LDA':
        rpath = '/pub/abinitio/Psps/LDA_GTH/'
    elif kind == 'CORE' and exchange == 'LDA':
        rpath = '/pub/abinitio/Psps/LDA_Core/'
    elif kind == 'HGH' and exchange == 'LDA':
        rpath = '/pub/abinitio/Psps/LDA_HGH/'
    elif kind == 'HGH' and exchange == 'GGA':
        rpath = '/pub/abinitio/Psps/GGA_HGH/'
    elif kind == 'TM' and exchange == 'LDA':
        if atomicnumber is None:
            raise ValueError('Atomic number required')
        rpath = '/pub/abinitio/Psps/LDA_TM.psps/' + str(atomicnumber).zfill(2) + '/'
    elif kind == 'AE' and exchange == 'DEN':
        rpath = '/pub/abinitio/Psps/AE_DEN/'
    elif kind == 'FC' and exchange == 'DEN':
        rpath = '/pub/abinitio/Psps/FC_DEN/'
    elif kind == 'PAW' and exchange == 'LDA':
        rpath = ' https://www.abinit.org/ATOMICDATA/JTH-LDA-atomicdata.tar.gz'
    elif kind == 'PAW' and exchange == 'PBE':
        rpath = ' https://www.abinit.org/ATOMICDATA/JTH-PBE-atomicdata.tar.gz'
    elif kind == 'ONC' and exchange == 'PBE':
        rpath = ' http://www.pseudo-dojo.org/pseudos/nc-sr_pbe_standard_psp8.tgz'
    else:
        print('Not know kind of PSP')
    return rpath


def get_all_psps(basedir, exchange, kind):
    directory = basedir + os.sep + exchange + '_' + kind
    if not os.path.isdir(directory):
        os.mkdir(directory)
    if kind in ['PAW', 'ONC']:
        rpath = get_rpath_psp(kind, exchange)
        filename = rpath.split('/')[-1]
        if not os.path.isfile(directory + '/' + filename):
            u = urlopen(rpath)
            f = open(directory + os.sep + filename, 'wb')
            meta = u.info()
            file_size = int(meta.get("Content-Length"))
            print("Downloading: %s Bytes: %s" % (filename, file_size))
            file_size_dl = 0
            block_sz = 8192
            while True:
                readed_buffer = u.read(block_sz)
                if not readed_buffer:
                    break
                file_size_dl += len(readed_buffer)
                f.write(readed_buffer)
                status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
                status += chr(8) * (len(status) + 1)
                print(status, end='')
            f.close()
            print('\n')
        else:
            print('All files are downloaded')

        try:
            tar = tarfile.open(directory + '/' + filename, 'r:gz')
            for item in tar:
                if not os.path.exists(item.name):
                    tar.extract(item, path=directory)
        except tarfile.ReadError:
            name = os.path.basename(filename)
            print(name[:name.rfind('.')], '<filename>')

    elif kind == 'HGH':
        while True:
            succeed = True
            ftp = ftplib.FTP('ftp.abinit.org')  # connect to host, default port
            ftp.login()  # user anonymous, passwd anonymous@
            ftp.cwd('pub/abinitio/Psps/LDA_HGH/')
            for filename in ftp.nlst():
                if not os.path.exists(directory + '/' + filename) or os.path.getsize(directory + '/' + filename) == 0:
                    print('Getting %s' % filename)
                    try:
                        ftp.retrbinary('RETR ' + filename, open(directory + '/' + filename, 'wb').write)
                    except ftplib.error_perm:
                        print('Failed to get ', filename)
                        succeed = False
            ftp.close()
            if succeed:
                print('All files are downloaded')
                break
    else:
        files = []
        for i in range(1, 113):
            if kind == 'GTH' and i > 17:
                continue
            if kind == 'CORE' and i not in [6, 7]:
                continue
            if kind == 'FHI' and exchange == 'LDA' and i in [57, 59, 63, 65, 66, 67, 90, 94, 102, 110, 111, 112]:
                continue
            if kind == 'TM' and exchange == 'LDA' and i in [104, 105, 106, 107, 108, 109, 110, 111, 112]:
                continue
            if kind == 'FHI' and exchange == 'GGA' and i in [57, 59, 63, 65, 66, 67, 90, 94, 102, 110, 111, 112]:
                continue
            if kind == 'AE' and exchange == 'DEN' and i in [57, 59, 63, 65, 66, 67, 90, 94, 102, 110, 111, 112]:
                continue
            if kind == 'GTH' and exchange == 'LDA' and i in [2, 10]:
                continue
            if kind == 'FC' and exchange == 'DEN' and i in [63, 65, 66, 67, 110, 111, 112]:
                continue
            filename = psp_name(i, exchange, kind)
            filepath = get_rpath_psp(kind, exchange, i)
            if not os.path.isfile(directory + '/' + filename) or os.path.getsize(directory + '/' + filename) == 0:
                files.append((filename, directory, filepath))

        if len(files) == 0:
            print("All files are downloaded")
        else:
            print("Downloading %d PSP files" % len(files))
        nth = 12
        pool = ThreadPoolExecutor(nth)
        index = 0
        p = nth * [None]
        while index < len(files):
            for i in range(nth):
                if index < len(files):
                    p[i] = pool.submit(worker, *(files[index]))
                    index += 1
            for i in range(nth):
                try:
                    p[i].result()
                except AttributeError:
                    print("Complete")

        if len(files) > 0:
            print("kind == '%s' and exchange == '%s'" % (kind, exchange))
            for i in files:
                print(" %s" % str(i))


if __name__ == '__main__':

    print("Script to download PSP files from ftp.abinit.org")
    home = os.environ['HOME']
    basedir = home + "/.abinit"

    print("Files will be downloaded at: %s/.abinit" % home)

    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    for exchange in ['LDA', 'GGA', 'DEN', 'PBE']:
        print('\n=> ' + exchange, end='\n')
        lista = []
        if exchange == 'LDA':
            lista = ['FHI', 'TM', 'GTH', 'PAW', 'CORE', 'HGH']
        elif exchange == 'GGA':
            lista = ['FHI', 'HGH']
        elif exchange == 'DEN':
            lista = ['AE', 'FC']
        elif exchange == 'PBE':
            lista = ['ONC', 'PAW']
        for kind in lista:
            print('--> %5s ' % kind, end=' ')
            get_all_psps(basedir, exchange, kind)
